# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import dataclasses
import json
import os
import re
import sys
from collections.abc import Callable, Sequence
from enum import Enum
from typing import TYPE_CHECKING, Any, Type, cast

from typing_extensions import TypedDict, Unpack

import anki
import anki.lang
from anki._legacy import deprecated
from anki.lang import is_rtl
from anki.utils import hmr_mode, is_lin, is_mac, is_win
from aqt import colors, gui_hooks
from aqt.qt import *
from aqt.qt import sip
from aqt.theme import theme_manager
from aqt.utils import askUser, is_gesture_or_zoom_event, openLink, showInfo, tr

serverbaseurl = re.compile(r"^.+:\/\/[^\/]+")

if TYPE_CHECKING:
    from aqt.mediasrv import PageContext

BridgeCommandHandler = Callable[[str], Any]


class AnkiWebViewKind(Enum):
    """Enum registry of all web views managed by Anki

    The value of each entry corresponds to the web view's title.

    When introducing a new web view, please add it to the registry below.
    """

    DEFAULT = "default"
    MAIN = "main webview"
    TOP_TOOLBAR = "top toolbar"
    BOTTOM_TOOLBAR = "bottom toolbar"
    DECK_OPTIONS = "deck options"
    EDITOR = "editor"
    LEGACY_DECK_STATS = "legacy deck stats"
    DECK_STATS = "deck stats"
    PREVIEWER = "previewer"
    CHANGE_NOTETYPE = "change notetype"
    CARD_LAYOUT = "card layout"
    BROWSER_CARD_INFO = "browser card info"
    IMPORT_CSV = "csv import"
    EMPTY_CARDS = "empty cards"
    FIND_DUPLICATES = "find duplicates"
    FIELDS = "fields"
    IMPORT_LOG = "import log"
    IMPORT_ANKI_PACKAGE = "anki package import"


class AuthInterceptor(QWebEngineUrlRequestInterceptor):
    _api_enabled = False

    def __init__(self, parent: QObject | None = None, api_enabled: bool = False):
        super().__init__(parent)
        self._api_enabled = api_enabled

    def interceptRequest(self, info):
        from aqt.mediasrv import _APIKEY

        if self._api_enabled and info.requestUrl().host() == "127.0.0.1":
            info.setHttpHeader(b"Authorization", f"Bearer {_APIKEY}".encode("utf-8"))


def _create_bridge_script() -> QWebEngineScript:
    qwebchannel = ":/qtwebchannel/qwebchannel.js"
    jsfile = QFile(qwebchannel)
    if not jsfile.open(QIODevice.OpenModeFlag.ReadOnly):
        print(f"Error opening '{qwebchannel}': {jsfile.error()}", file=sys.stderr)
    jstext = bytes(cast(bytes, jsfile.readAll())).decode("utf-8")
    jsfile.close()

    script = QWebEngineScript()
    script.setSourceCode(
        jstext
        + """
        var pycmd, bridgeCommand;
        new QWebChannel(qt.webChannelTransport, function(channel) {
            bridgeCommand = pycmd = function (arg, cb) {
                var resultCB = function (res) {
                    // pass result back to user-provided callback
                    if (cb) {
                        cb(JSON.parse(res));
                    }
                }
            
                channel.objects.py.cmd(arg, resultCB);
                return false;                   
            }
            pycmd("domDone");
        });
    """
    )
    script.setWorldId(QWebEngineScript.ScriptWorldId.MainWorld)
    script.setInjectionPoint(QWebEngineScript.InjectionPoint.DocumentReady)
    script.setRunsOnSubFrames(False)

    return script


_bridge_script = _create_bridge_script()

_profile_with_api_access: QWebEngineProfile | None = None
_profile_without_api_access: QWebEngineProfile | None = None


class AnkiWebPage(QWebEnginePage):
    def __init__(
        self,
        onBridgeCmd: BridgeCommandHandler,
        kind: AnkiWebViewKind = AnkiWebViewKind.DEFAULT,
        parent: QObject | None = None,
    ) -> None:
        profile = self._profileForPage(kind)
        self._inject_user_script(profile, _bridge_script)
        QWebEnginePage.__init__(self, profile, parent)
        self._onBridgeCmd = onBridgeCmd
        self._kind = kind
        self._setupBridge()
        self.open_links_externally = True

    def _profileForPage(self, kind: AnkiWebViewKind) -> QWebEngineProfile:
        have_api_access = kind in (
            AnkiWebViewKind.DECK_OPTIONS,
            AnkiWebViewKind.EDITOR,
            AnkiWebViewKind.DECK_STATS,
            AnkiWebViewKind.CHANGE_NOTETYPE,
            AnkiWebViewKind.BROWSER_CARD_INFO,
            AnkiWebViewKind.IMPORT_ANKI_PACKAGE,
            AnkiWebViewKind.IMPORT_CSV,
            AnkiWebViewKind.IMPORT_LOG,
        )

        global _profile_with_api_access, _profile_without_api_access

        # Use cached profile if available
        if have_api_access and _profile_with_api_access is not None:
            return _profile_with_api_access
        elif not have_api_access and _profile_without_api_access is not None:
            return _profile_without_api_access

        # Create a new profile if not cached
        profile = QWebEngineProfile()

        interceptor = AuthInterceptor(profile, api_enabled=have_api_access)
        profile.setUrlRequestInterceptor(interceptor)
        if have_api_access:
            _profile_with_api_access = profile
        else:
            _profile_without_api_access = profile

        return profile

    def _setupBridge(self) -> None:
        # Add-on compatibility: For existing add-on callers that override the init
        # and invoke _setupBridge directly (e.g. in order to use a custom web profile),
        # we need to ensure that the bridge script is injected into the profile scripts,
        # if it has yet to be injected.
        profile = self.profile()
        assert profile is not None
        scripts = profile.scripts()
        assert scripts is not None

        if not scripts.contains(_bridge_script):
            print("add-on callers should not call _setupBridge directly")
            self._inject_user_script(profile, _bridge_script)

        class Bridge(QObject):
            def __init__(self, bridge_handler: Callable[[str], Any]) -> None:
                super().__init__()
                self.onCmd = bridge_handler

            @pyqtSlot(str, result=str)  # type: ignore
            def cmd(self, str: str) -> Any:
                return json.dumps(self.onCmd(str))

        self._bridge = Bridge(self._onCmd)

        self._channel = QWebChannel(self)
        self._channel.registerObject("py", self._bridge)
        self.setWebChannel(self._channel)

    def _inject_user_script(
        self, profile: QWebEngineProfile, script: QWebEngineScript
    ) -> None:
        scripts = profile.scripts()
        assert scripts is not None
        scripts.insert(script)

    def javaScriptConsoleMessage(
        self,
        level: QWebEnginePage.JavaScriptConsoleMessageLevel,
        msg: str | None,
        line: int,
        srcID: str | None,
    ) -> None:
        # not translated because console usually not visible,
        # and may only accept ascii text
        assert srcID is not None
        if srcID.startswith("data"):
            srcID = ""
        else:
            srcID = serverbaseurl.sub("", srcID[:80], 1)
        if level == QWebEnginePage.JavaScriptConsoleMessageLevel.InfoMessageLevel:
            level_str = "info"
        elif level == QWebEnginePage.JavaScriptConsoleMessageLevel.WarningMessageLevel:
            level_str = "warning"
        elif level == QWebEnginePage.JavaScriptConsoleMessageLevel.ErrorMessageLevel:
            level_str = "error"
        else:
            level_str = str(level)
        buf = "JS %(t)s %(f)s:%(a)d %(b)s" % dict(
            t=level_str, a=line, f=srcID, b=f"{msg}\n"
        )
        if "MathJax localStorage" in buf:
            # silence localStorage noise
            return
        elif "link preload" in buf:
            # silence 'link preload' warning on the first card
            return
        # ensure we don't try to write characters the terminal can't handle
        buf = buf.encode(sys.stdout.encoding, "backslashreplace").decode(
            sys.stdout.encoding
        )
        # output to stdout because it may raise error messages on the anki GUI
        # https://github.com/ankitects/anki/pull/560
        sys.stdout.write(buf)

    def acceptNavigationRequest(
        self, url: QUrl, navType: Any, isMainFrame: bool
    ) -> bool:
        from aqt.mediasrv import is_sveltekit_page

        if (
            not self.open_links_externally
            or "_anki/pages" in url.path()
            or url.path() == "/_anki/legacyPageData"
            or is_sveltekit_page(url.path()[1:])
        ):
            return super().acceptNavigationRequest(url, navType, isMainFrame)

        if not isMainFrame:
            return True
        # data: links generated by setHtml()
        if url.scheme() == "data":
            return True
        # catch buggy <a href='#' onclick='func()'> links
        from aqt import mw

        if url.matches(
            QUrl(mw.serverURL()), cast(Any, QUrl.UrlFormattingOption.RemoveFragment)
        ):
            print("onclick handler needs to return false")
            return False
        # load all other links in browser
        from aqt.url_schemes import open_url_if_supported_scheme

        open_url_if_supported_scheme(url)
        return False

    def _onCmd(self, str: str) -> Any:
        return self._onBridgeCmd(str)

    def javaScriptAlert(self, frame: Any, text: str | None) -> None:
        if text is None:
            return

        showInfo(text)

    def javaScriptConfirm(self, frame: Any, text: str | None) -> bool:
        if text is None:
            return False

        return askUser(text)


# Add-ons
##########################################################################


@dataclasses.dataclass
class WebContent:
    """Stores all dynamically modified content that a particular web view
    will be populated with.

    Attributes:
        body {str} -- HTML body
        head {str} -- HTML head
        css {List[str]} -- List of media server subpaths,
                           each pointing to a CSS file
        js {List[str]} -- List of media server subpaths,
                          each pointing to a JS file

    Important Notes:
        - When modifying the attributes specified above, please make sure your
        changes only perform the minimum required edits to make your add-on work.
        You should avoid overwriting or interfering with existing data as much
        as possible, instead opting to append your own changes, e.g.:

            def on_webview_will_set_content(web_content: WebContent, context) -> None:
                web_content.body += "<my_html>"
                web_content.head += "<my_head>"

        - The paths specified in `css` and `js` need to be accessible by Anki's
          media server. All list members without a specified subpath are assumed
          to be located under `/_anki`, which is the media server subpath used
          for all web assets shipped with Anki.

          Add-ons may expose their own web assets by utilizing
          aqt.addons.AddonManager.setWebExports(). Web exports registered
          in this manner may then be accessed under the `/_addons` subpath.

          E.g., to allow access to a `my-addon.js` and `my-addon.css` residing
          in a "web" subfolder in your add-on package, first register the
          corresponding web export:

          > from aqt import mw
          > mw.addonManager.setWebExports(__name__, r"web/.*(css|js)")

          Then append the subpaths to the corresponding web_content fields
          within a function subscribing to gui_hooks.webview_will_set_content:

              def on_webview_will_set_content(web_content: WebContent, context) -> None:
                  addon_package = mw.addonManager.addonFromModule(__name__)
                  web_content.css.append(
                      f"/_addons/{addon_package}/web/my-addon.css")
                  web_content.js.append(
                      f"/_addons/{addon_package}/web/my-addon.js")

          Note that '/' will also match the os specific path separator.
    """

    body: str = ""
    head: str = ""
    css: list[str] = dataclasses.field(default_factory=lambda: [])
    js: list[str] = dataclasses.field(default_factory=lambda: [])


# Main web view
##########################################################################


class AnkiWebView(QWebEngineView):
    allow_drops = False
    _kind: AnkiWebViewKind

    def __init__(
        self,
        parent: QWidget | None = None,
        title: str = "",  # used by add-ons; in Anki code use kind instead to set title
        kind: AnkiWebViewKind = AnkiWebViewKind.DEFAULT,
    ) -> None:
        QWebEngineView.__init__(self, parent=parent)
        self._kind = kind
        self.set_title(kind.value)
        self.setPage(AnkiWebPage(self._onBridgeCmd, kind, self))
        # reduce flicker
        self.page().setBackgroundColor(theme_manager.qcolor(colors.CANVAS))

        # in new code, use .set_bridge_command() instead of setting this directly
        self.onBridgeCmd: Callable[[str], Any] = self.defaultOnBridgeCmd

        self._domDone = True
        self._pendingActions: list[tuple[str, Sequence[Any]]] = []
        self.requiresCol = True
        self._disable_zoom = False

        self.resetHandlers()
        self._filterSet = False
        gui_hooks.theme_did_change.append(self.on_theme_did_change)
        gui_hooks.body_classes_need_update.append(self.on_body_classes_need_update)

        qconnect(self.loadFinished, self._on_load_finished)

    def _on_load_finished(self) -> None:
        self.eval(
            """
        document.addEventListener("keydown", function(evt) {
            if (evt.key === "Escape") {
                pycmd("close");
            }
        });
        """
        )

    def page(self) -> AnkiWebPage:
        return cast(AnkiWebPage, super().page())

    @property
    def kind(self) -> AnkiWebViewKind:
        """Used by add-ons to identify the webview kind"""
        return self._kind

    def set_title(self, title: str) -> None:
        self.title = title  # type: ignore[assignment]

    def disable_zoom(self) -> None:
        self._disable_zoom = True

    def createWindow(self, windowType: QWebEnginePage.WebWindowType) -> QWebEngineView:
        # intercept opening a new window (hrefs
        # with target="_blank") and return view
        return AnkiWebView()

    def eventFilter(self, obj: QObject | None, evt: QEvent | None) -> bool:
        if evt is None:
            return False
        if self._disable_zoom and is_gesture_or_zoom_event(evt):
            return True

        if (
            isinstance(evt, QMouseEvent)
            and evt.type() == QEvent.Type.MouseButtonRelease
        ):
            from aqt import mw

            if evt.button() == Qt.MouseButton.MiddleButton and is_lin:
                if mw.pm.middle_click_paste_enabled():
                    self.onMiddleClickPaste()
                return True

        return False

    def set_open_links_externally(self, enable: bool) -> None:
        self.page().open_links_externally = enable

    def onEsc(self) -> None:
        w = self.parent()
        while w:
            if isinstance(w, QDialog) or isinstance(w, QMainWindow):
                from aqt import mw

                # esc in a child window closes the window
                if w != mw:
                    w.close()
                else:
                    # in the main window, removes focus from type in area
                    parent = self.parent()
                    assert isinstance(parent, QWidget)
                    parent.setFocus()
                break
            w = w.parent()

    def onCopy(self) -> None:
        self.triggerPageAction(QWebEnginePage.WebAction.Copy)

    def onCut(self) -> None:
        self.triggerPageAction(QWebEnginePage.WebAction.Cut)

    def onPaste(self) -> None:
        self.triggerPageAction(QWebEnginePage.WebAction.Paste)

    def onMiddleClickPaste(self) -> None:
        self.triggerPageAction(QWebEnginePage.WebAction.Paste)

    def onSelectAll(self) -> None:
        self.triggerPageAction(QWebEnginePage.WebAction.SelectAll)

    def contextMenuEvent(self, evt: QContextMenuEvent | None) -> None:
        m = QMenu(self)
        self._maybe_add_copy_action(m)
        gui_hooks.webview_will_show_context_menu(self, m)
        if m.actions():
            m.popup(QCursor.pos())

    def _maybe_add_copy_action(self, menu: QMenu) -> None:
        if self.hasSelection():
            a = menu.addAction(tr.actions_copy())
            assert a is not None
            qconnect(a.triggered, self.onCopy)

    def dropEvent(self, evt: QDropEvent | None) -> None:
        if self.allow_drops:
            super().dropEvent(evt)

    def setHtml(  #  type: ignore[override]
        self, html: str, context: PageContext | None = None
    ) -> None:
        from aqt.mediasrv import PageContext

        # discard any previous pending actions
        self._pendingActions = []
        self._domDone = True
        if context is None:
            context = PageContext.UNKNOWN
        self._queueAction("setHtml", html, context)
        self.set_open_links_externally(True)
        self.allow_drops = False
        self.show()

    def _setHtml(self, html: str, context: PageContext) -> None:
        """Send page data to media server, then surf to it.

        This function used to be implemented by QWebEngine's
        .setHtml() call. It is no longer used, as it has a
        maximum size limit, and due to security changes, it
        will stop working in the future."""
        from aqt import mw

        oldFocus = mw.app.focusWidget()
        self._domDone = False

        webview_id = id(self)
        mw.mediaServer.set_page_html(webview_id, html, context)
        self.load_url(QUrl(f"{mw.serverURL()}_anki/legacyPageData?id={webview_id}"))

        # work around webengine stealing focus on setHtml()
        # fixme: check which if any qt versions this is still required on
        if oldFocus:
            oldFocus.setFocus()

    def load_url(self, url: QUrl) -> None:
        # allow queuing actions when loading url directly
        self._domDone = False
        self.allow_drops = False
        super().load(url)

    def app_zoom_factor(self) -> float:
        # overridden scale factor?
        webscale = os.environ.get("ANKI_WEBSCALE")
        if webscale:
            return float(webscale)

        if qtmajor > 5 or is_mac:
            return 1
        screen = QApplication.desktop().screen()  # type: ignore
        if screen is None:
            return 1

        dpi = screen.logicalDpiX()
        factor = dpi / 96.0
        if is_lin:
            factor = max(1, factor)
            return factor
        return 1

    def setPlaybackRequiresGesture(self, value: bool) -> None:
        settings = self.settings()
        assert settings is not None
        settings.setAttribute(
            QWebEngineSettings.WebAttribute.PlaybackRequiresUserGesture, value
        )

    def _getQtIntScale(self, screen: QWidget) -> int:
        # try to detect if Qt has scaled the screen
        # - qt will round the scale factor to a whole number, so a dpi of 125% = 1x,
        #   and a dpi of 150% = 2x
        # - a screen with a normal physical dpi of 72 will have a dpi of 32
        #   if the scale factor has been rounded to 2x
        # - different screens have different physical DPIs (eg 72, 93, 102)
        # - until a better solution presents itself, assume a physical DPI at
        #   or above 70 is unscaled
        if screen.physicalDpiX() > 70:
            return 1
        elif screen.physicalDpiX() > 35:
            return 2
        else:
            return 3

    def standard_css(self) -> str:
        color_hl = theme_manager.var(colors.BORDER_FOCUS)

        if is_win:
            # T: include a font for your language on Windows, eg: "Segoe UI", "MS Mincho"
            family = tr.qt_misc_segoe_ui()
            button_style = f"""
button {{ font-family: {family}; }}
            """
            font = f"font-family:{family};"
        elif is_mac:
            family = "Helvetica"
            font = f'font-family:"{family}";'
            button_style = """
button {
    --canvas: #fff;
    -webkit-appearance: none;
    background: var(--canvas);
    border-radius: var(--border-radius);
    padding: 3px 12px;
    border: 1px solid var(--border);
    box-shadow: 0px 1px 3px var(--border-subtle);
    font-family: Helvetica
}
.night-mode button { --canvas: #606060; --fg: #eee; }
"""
        else:
            family = self.font().family()
            font = f'font-family:"{family}", sans-serif;'
            button_style = """
/* Buttons */
button{{ 
    font-family: "{family}", sans-serif;
}}
/* Input field focus outline */
textarea:focus, input:focus, input[type]:focus, .uneditable-input:focus,
div[contenteditable="true"]:focus {{   
    outline: 0 none;
    border-color: {color_hl};
}}""".format(
                family=family,
                color_hl=color_hl,
            )

        zoom = self.app_zoom_factor()

        return f"""
body {{ zoom: {zoom}; background-color: var(--canvas); }}
html {{ {font} }}
{button_style}
:root {{ --canvas: {colors.CANVAS["light"]} }}
:root[class*=night-mode] {{ --canvas: {colors.CANVAS["dark"]} }}
"""

    def stdHtml(
        self,
        body: str,
        css: list[str] | None = None,
        js: list[str] | None = None,
        head: str = "",
        context: Any | None = None,
        default_css: bool = True,
    ) -> None:
        css = (["css/webview.css"] if default_css else []) + (
            [] if css is None else css
        )
        web_content = WebContent(
            body=body,
            head=head,
            js=["js/webview.js"] + (["js/vendor/jquery.min.js"] if js is None else js),
            css=css,
        )

        gui_hooks.webview_will_set_content(web_content, context)

        csstxt = ""
        if "css/webview.css" in css:
            # we want our dynamic styling to override the defaults in
            # css/webview.css, but come before user-provided stylesheets so that
            # they can override us if necessary
            web_content.css.remove("css/webview.css")
            csstxt = self.bundledCSS("css/webview.css")
            csstxt += f"<style>{self.standard_css()}</style>"

        csstxt += "\n".join(self.bundledCSS(fname) for fname in web_content.css)
        jstxt = "\n".join(self.bundledScript(fname) for fname in web_content.js)

        from aqt import mw

        head = mw.baseHTML() + csstxt + web_content.head
        body_class = theme_manager.body_class()

        if theme_manager.night_mode:
            doc_class = "night-mode"
            bs_theme = "dark"
        else:
            doc_class = ""
            bs_theme = "light"

        if is_rtl(anki.lang.current_lang):
            lang_dir = "rtl"
        else:
            lang_dir = "ltr"

        html = f"""
<!doctype html>
<html class="{doc_class}" dir="{lang_dir}" data-bs-theme="{bs_theme}">
<head>
    <title>{self.title}</title>
{head}
</head>

<body class="{body_class}">
{jstxt}
{web_content.body}</body>
</html>"""
        # print(html)
        import aqt.browser.previewer
        import aqt.clayout
        import aqt.deckoptions
        import aqt.editor
        import aqt.reviewer
        from aqt.mediasrv import PageContext

        if isinstance(context, aqt.editor.Editor):
            page_context = PageContext.EDITOR
        elif isinstance(context, aqt.reviewer.Reviewer):
            page_context = PageContext.REVIEWER
        elif isinstance(context, aqt.browser.previewer.Previewer):
            page_context = PageContext.PREVIEWER
        elif isinstance(context, aqt.clayout.CardLayout):
            page_context = PageContext.CARD_LAYOUT
        elif isinstance(context, aqt.deckoptions.DeckOptionsDialog):
            page_context = PageContext.DECK_OPTIONS
        else:
            page_context = PageContext.UNKNOWN
        self.setHtml(html, page_context)

    @classmethod
    def webBundlePath(cls, path: str) -> str:
        from aqt import mw

        if path.startswith("/"):
            subpath = ""
        else:
            subpath = "/_anki/"

        return f"http://127.0.0.1:{mw.mediaServer.getPort()}{subpath}{path}"

    def bundledScript(self, fname: str) -> str:
        return f'<script src="{self.webBundlePath(fname)}"></script>'

    def bundledCSS(self, fname: str) -> str:
        return '<link rel="stylesheet" type="text/css" href="%s">' % self.webBundlePath(
            fname
        )

    def eval(self, js: str) -> None:
        self.evalWithCallback(js, None)

    def evalWithCallback(self, js: str, cb: Callable | None) -> None:
        self._queueAction("eval", js, cb)

    def _evalWithCallback(self, js: str, cb: Callable[[Any], Any] | None) -> None:
        page = self.page()
        assert page is not None

        def handler(val: Any) -> None:
            if self._shouldIgnoreWebEvent():
                print("ignored late js callback", cb)
                return
            if cb:
                cb(val)

            # Without the following, stale frames showing previous or corrupt content get occasionally displayed. (see #3668 for more details)
            self.update()

        page.runJavaScript(js, handler)

    def _queueAction(self, name: str, *args: Any) -> None:
        self._pendingActions.append((name, args))
        self._maybeRunActions()

    def _maybeRunActions(self) -> None:
        if sip.isdeleted(self):
            return
        while self._pendingActions and self._domDone:
            name, args = self._pendingActions.pop(0)

            if name == "eval":
                self._evalWithCallback(*args)
            elif name == "setHtml":
                self._setHtml(*args)
            else:
                raise Exception(f"unknown action: {name}")

    def _openLinksExternally(self, url: str) -> None:
        openLink(url)

    def _shouldIgnoreWebEvent(self) -> bool:
        # async web events may be received after the profile has been closed
        # or the underlying webview has been deleted
        from aqt import mw

        if sip.isdeleted(self):
            return True
        if not mw.col and self.requiresCol:
            return True
        return False

    def _onBridgeCmd(self, cmd: str) -> Any:
        if self._shouldIgnoreWebEvent():
            print("ignored late bridge cmd", cmd)
            return

        if not self._filterSet:
            focus_proxy = self.focusProxy()
            assert focus_proxy is not None
            focus_proxy.installEventFilter(self)
            self._filterSet = True

        if cmd == "domDone":
            self._domDone = True
            self._maybeRunActions()
        elif cmd == "close":
            self.onEsc()
        else:
            handled, result = gui_hooks.webview_did_receive_js_message(
                (False, None), cmd, self._bridge_context
            )
            if handled:
                return result
            else:
                return self.onBridgeCmd(cmd)

    def defaultOnBridgeCmd(self, cmd: str) -> None:
        print("unhandled bridge cmd:", cmd)

    # legacy
    def resetHandlers(self) -> None:
        self.onBridgeCmd = self.defaultOnBridgeCmd
        self._bridge_context = None

    def adjustHeightToFit(self) -> None:
        self.evalWithCallback("document.documentElement.offsetHeight", self._onHeight)

    def _onHeight(self, qvar: int | None) -> None:
        from aqt import mw

        if qvar is None:
            mw.progress.single_shot(1000, mw.reset)
            return

        self.setFixedHeight(int(qvar))

    def set_bridge_command(self, func: Callable[[str], Any], context: Any) -> None:
        """Set a handler for pycmd() messages received from Javascript.

        Context is the object calling this routine, eg an instance of
        aqt.reviewer.Reviewer or aqt.deckbrowser.DeckBrowser."""
        self.onBridgeCmd = func
        self._bridge_context = context

    def hide_while_preserving_layout(self) -> None:
        "Hide but keep existing size."
        sp = self.sizePolicy()
        sp.setRetainSizeWhenHidden(True)
        self.setSizePolicy(sp)
        self.hide()

    def add_dynamic_styling_and_props_then_show(self) -> None:
        "Add dynamic styling, title, set platform-specific body classes and reveal."
        css = self.standard_css()
        body_classes = theme_manager.body_class().split(" ")

        def after_injection(arg: Any) -> None:
            gui_hooks.webview_did_inject_style_into_page(self)
            self.show()

        if theme_manager.night_mode:
            night_mode = 'document.documentElement.classList.add("night-mode");'
        else:
            night_mode = ""
        self.evalWithCallback(
            f"""
(function(){{
    document.title = `{self.title}`;
    const style = document.createElement('style');
    style.innerHTML = `{css}`;
    document.head.appendChild(style);
    document.body.classList.add({", ".join([f'"{c}"' for c in body_classes])});
    {night_mode}
}})();
""",
            after_injection,
        )

    def load_ts_page(self, name: str) -> None:
        from aqt import mw

        self.set_open_links_externally(True)
        if theme_manager.night_mode:
            extra = "#night"
        else:
            extra = ""
        self.load_url(QUrl(f"{mw.serverURL()}_anki/pages/{name}.html{extra}"))
        self.add_dynamic_styling_and_props_then_show()

    def load_sveltekit_page(self, path: str) -> None:
        from aqt import mw

        self.set_open_links_externally(True)
        if theme_manager.night_mode:
            extra = "#night"
        else:
            extra = ""

        if hmr_mode:
            server = "http://127.0.0.1:5173/"
        else:
            server = mw.serverURL()

        self.load_url(QUrl(f"{server}{path}{extra}"))
        self.add_dynamic_styling_and_props_then_show()

    def force_load_hack(self) -> None:
        """Force process to initialize.
        Must be done on Windows prior to changing current working directory."""
        self.requiresCol = False
        self._domReady = False
        self.page().setContent(cast(QByteArray, bytes("", "ascii")))

    def cleanup(self) -> None:
        try:
            from aqt import mw
        except ImportError:
            # this will fail when __del__ is called during app shutdown
            return

        gui_hooks.theme_did_change.remove(self.on_theme_did_change)
        gui_hooks.body_classes_need_update.remove(self.on_body_classes_need_update)
        # defer page cleanup so that in-flight requests have a chance to complete first
        # https://forums.ankiweb.net/t/error-when-exiting-browsing-when-the-software-is-installed-in-the-path-c-program-files-anki/38363
        mw.progress.single_shot(5000, lambda: mw.mediaServer.clear_page_html(id(self)))
        self.page().deleteLater()

    def on_theme_did_change(self) -> None:
        # avoid flashes if page reloaded
        self.page().setBackgroundColor(theme_manager.qcolor(colors.CANVAS))
        if hasattr(QWebEngineSettings.WebAttribute, "ForceDarkMode"):
            force_dark_mode = getattr(QWebEngineSettings.WebAttribute, "ForceDarkMode")
            page_settings = self.page().settings()
            if page_settings is not None:
                page_settings.setAttribute(
                    force_dark_mode,
                    theme_manager.get_night_mode(),
                )
        # update night-mode class, and legacy nightMode/night-mode body classes
        self.eval(
            f"""
(function() {{
    const doc = document.documentElement;
    const body = document.body.classList;
    if ({1 if theme_manager.night_mode else 0}) {{
        doc.dataset.bsTheme = "dark";
        doc.classList.add("night-mode");
        body.add("night_mode");
        body.add("nightMode");
        {"body.add('macos-dark-mode');" if theme_manager.macos_dark_mode() else ""}
    }} else {{
        doc.dataset.bsTheme = "light";
        doc.classList.remove("night-mode");
        body.remove("night_mode");
        body.remove("nightMode");
        body.remove("macos-dark-mode");
    }}
}})();
"""
        )

    def on_body_classes_need_update(self) -> None:
        from aqt import mw

        self.eval(
            f"""document.body.classList.toggle("fancy", {json.dumps(not mw.pm.minimalist_mode())}); """
        )
        self.eval(
            f"""document.body.classList.toggle("reduce-motion", {json.dumps(mw.pm.reduce_motion())}); """
        )

    @deprecated(info="use theme_manager.qcolor() instead")
    def get_window_bg_color(self, night_mode: bool | None = None) -> QColor:
        return theme_manager.qcolor(colors.CANVAS)


# Pre-configured classes for use in Qt Designer
##########################################################################


class _AnkiWebViewKwargs(TypedDict, total=False):
    parent: QWidget | None
    title: str
    kind: AnkiWebViewKind


def _create_ankiwebview_subclass(
    name: str,
    /,
    **fixed_kwargs: Unpack[_AnkiWebViewKwargs],
) -> Type[AnkiWebView]:
    def __init__(self, *args: Any, **kwargs: _AnkiWebViewKwargs) -> None:
        # user‑supplied kwargs override fixed kwargs
        merged = cast(_AnkiWebViewKwargs, {**fixed_kwargs, **kwargs})
        AnkiWebView.__init__(self, *args, **merged)

    __init__.__qualname__ = f"{name}.__init__"
    if fixed_kwargs:
        __init__.__doc__ = (
            f"Auto‑generated wrapper that pre‑sets "
            f"{', '.join(f'{k}={v!r}' for k, v in fixed_kwargs.items())}."
        )

    cls: Type[AnkiWebView] = type(name, (AnkiWebView,), {"__init__": __init__})

    return cls


# These subclasses are used in Qt Designer UI files to allow for configuring
# web views at initialization time (custom widgets can otherwise only be
# initialized with the default constructor)
StatsWebView = _create_ankiwebview_subclass(
    "StatsWebView", kind=AnkiWebViewKind.DECK_STATS
)
LegacyStatsWebView = _create_ankiwebview_subclass(
    "LegacyStatsWebView", kind=AnkiWebViewKind.LEGACY_DECK_STATS
)
EmptyCardsWebView = _create_ankiwebview_subclass(
    "EmptyCardsWebView", kind=AnkiWebViewKind.EMPTY_CARDS
)
FindDupesWebView = _create_ankiwebview_subclass(
    "FindDupesWebView", kind=AnkiWebViewKind.FIND_DUPLICATES
)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

import json
import os
import re
import shutil
import threading
import time
import unicodedata
import zipfile
from collections.abc import Sequence
from io import BufferedWriter
from typing import Any
from zipfile import ZipFile

from anki import hooks
from anki.cards import CardId
from anki.collection import Collection
from anki.decks import DeckId
from anki.utils import ids2str, namedtmp, split_fields, strip_html


class Exporter:
    includeHTML: bool | None = None
    ext: str | None = None
    includeTags: bool | None = None
    includeSched: bool | None = None
    includeMedia: bool | None = None

    def __init__(
        self,
        col: Collection,
        did: DeckId | None = None,
        cids: list[CardId] | None = None,
    ) -> None:
        self.col = col.weakref()
        self.did = did
        self.cids = cids

    @staticmethod
    def key(col: Collection) -> str:
        return ""

    def doExport(self, path) -> None:
        raise Exception("not implemented")

    def exportInto(self, path: str) -> None:
        self._escapeCount = 0
        file = open(path, "wb")
        self.doExport(file)
        file.close()

    def processText(self, text: str) -> str:
        if self.includeHTML is False:
            text = self.stripHTML(text)

        text = self.escapeText(text)

        return text

    def escapeText(self, text: str) -> str:
        "Escape newlines, tabs, CSS and quotechar."
        # fixme: we should probably quote fields with newlines
        # instead of converting them to spaces
        text = text.replace("\n", " ")
        text = text.replace("\r", "")
        text = text.replace("\t", " " * 8)
        text = re.sub("(?i)<style>.*?</style>", "", text)
        text = re.sub(r"\[\[type:[^]]+\]\]", "", text)
        if '"' in text or "'" in text:
            text = '"' + text.replace('"', '""') + '"'
        return text

    def stripHTML(self, text: str) -> str:
        # very basic conversion to text
        s = text
        s = re.sub(r"(?i)<(br ?/?|div|p)>", " ", s)
        s = re.sub(r"\[sound:[^]]+\]", "", s)
        s = strip_html(s)
        s = re.sub(r"[ \n\t]+", " ", s)
        s = s.strip()
        return s

    def cardIds(self) -> Any:
        if self.cids is not None:
            cids = self.cids
        elif not self.did:
            cids = self.col.db.list("select id from cards")
        else:
            cids = self.col.decks.cids(self.did, children=True)
        self.count = len(cids)
        return cids


# Cards as TSV
######################################################################


class TextCardExporter(Exporter):
    ext = ".txt"
    includeHTML = True

    def __init__(self, col) -> None:
        Exporter.__init__(self, col)

    @staticmethod
    def key(col: Collection) -> str:
        return col.tr.exporting_cards_in_plain_text()

    def doExport(self, file) -> None:
        ids = sorted(self.cardIds())
        strids = ids2str(ids)

        def esc(s):
            # strip off the repeated question in answer if exists
            s = re.sub("(?si)^.*<hr id=answer>\n*", "", s)
            return self.processText(s)

        out = ""
        for cid in ids:
            c = self.col.get_card(cid)
            out += esc(c.question())
            out += "\t" + esc(c.answer()) + "\n"
        file.write(out.encode("utf-8"))


# Notes as TSV
######################################################################


class TextNoteExporter(Exporter):
    ext = ".txt"
    includeTags = True
    includeHTML = True

    def __init__(self, col: Collection) -> None:
        Exporter.__init__(self, col)
        self.includeID = False

    @staticmethod
    def key(col: Collection) -> str:
        return col.tr.exporting_notes_in_plain_text()

    def doExport(self, file: BufferedWriter) -> None:
        cardIds = self.cardIds()
        data = []
        for id, flds, tags in self.col.db.execute(
            """
select guid, flds, tags from notes
where id in
(select nid from cards
where cards.id in %s)"""
            % ids2str(cardIds)
        ):
            row = []
            # note id
            if self.includeID:
                row.append(str(id))
            # fields
            row.extend([self.processText(f) for f in split_fields(flds)])
            # tags
            if self.includeTags:
                row.append(tags.strip())
            data.append("\t".join(row))
        self.count = len(data)
        out = "\n".join(data)
        file.write(out.encode("utf-8"))


# Anki decks
######################################################################
# media files are stored in self.mediaFiles, but not exported.


class AnkiExporter(Exporter):
    ext = ".anki2"
    includeSched: bool | None = False
    includeMedia = True

    def __init__(self, col: Collection) -> None:
        Exporter.__init__(self, col)

    @staticmethod
    def key(col: Collection) -> str:
        return col.tr.exporting_anki_20_deck()

    def deckIds(self) -> list[DeckId]:
        if self.cids:
            return self.col.decks.for_card_ids(self.cids)
        elif self.did:
            return self.src.decks.deck_and_child_ids(self.did)
        else:
            return []

    def exportInto(self, path: str) -> None:
        # create a new collection at the target
        try:
            os.unlink(path)
        except OSError:
            pass
        self.dst = Collection(path)
        self.src = self.col
        # find cards
        cids = self.cardIds()
        # copy cards, noting used nids
        nids = {}
        data: list[Sequence] = []
        for row in self.src.db.execute(
            "select * from cards where id in " + ids2str(cids)
        ):
            # clear flags
            row = list(row)
            row[-2] = 0
            nids[row[1]] = True
            data.append(row)
        self.dst.db.executemany(
            "insert into cards values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)", data
        )
        # notes
        strnids = ids2str(list(nids.keys()))
        notedata = []
        for row in self.src.db.all("select * from notes where id in " + strnids):
            # remove system tags if not exporting scheduling info
            if not self.includeSched:
                row = list(row)
                row[5] = self.removeSystemTags(row[5])
            notedata.append(row)
        self.dst.db.executemany(
            "insert into notes values (?,?,?,?,?,?,?,?,?,?,?)", notedata
        )
        # models used by the notes
        mids = self.dst.db.list("select distinct mid from notes where id in " + strnids)
        # card history and revlog
        if self.includeSched:
            data = self.src.db.all("select * from revlog where cid in " + ids2str(cids))
            self.dst.db.executemany(
                "insert into revlog values (?,?,?,?,?,?,?,?,?)", data
            )
        else:
            # need to reset card state
            self.dst.sched.reset_cards(cids)
        # models - start with zero
        self.dst.mod_schema(check=False)
        self.dst.models.remove_all_notetypes()
        for m in self.src.models.all():
            if int(m["id"]) in mids:
                self.dst.models.update(m)
        # decks
        dids = self.deckIds()
        dconfs = {}
        for d in self.src.decks.all():
            if str(d["id"]) == "1":
                continue
            if dids and d["id"] not in dids:
                continue
            if not d["dyn"] and d["conf"] != 1:
                if self.includeSched:
                    dconfs[d["conf"]] = True
            if not self.includeSched:
                # scheduling not included, so reset deck settings to default
                d = dict(d)
                d["conf"] = 1
                d["reviewLimit"] = d["newLimit"] = None
                d["reviewLimitToday"] = d["newLimitToday"] = None
            self.dst.decks.update(d)
        # copy used deck confs
        for dc in self.src.decks.all_config():
            if dc["id"] in dconfs:
                self.dst.decks.update_config(dc)
        # find used media
        media = {}
        self.mediaDir = self.src.media.dir()
        if self.includeMedia:
            for row in notedata:
                flds = row[6]
                mid = row[2]
                for file in self.src.media.files_in_str(mid, flds):
                    # skip files in subdirs
                    if file != os.path.basename(file):
                        continue
                    media[file] = True
            if self.mediaDir:
                for fname in os.listdir(self.mediaDir):
                    path = os.path.join(self.mediaDir, fname)
                    if os.path.isdir(path):
                        continue
                    if fname.startswith("_"):
                        # Scan all models in mids for reference to fname
                        for m in self.src.models.all():
                            if int(m["id"]) in mids:
                                if self._modelHasMedia(m, fname):
                                    media[fname] = True
                                    break
        self.mediaFiles = list(media.keys())
        self.dst.crt = self.src.crt
        # todo: tags?
        self.count = self.dst.card_count()
        self.postExport()
        self.dst.close(downgrade=True)

    def postExport(self) -> None:
        # overwrite to apply customizations to the deck before it's closed,
        # such as update the deck description
        pass

    def removeSystemTags(self, tags: str) -> str:
        return self.src.tags.rem_from_str("marked leech", tags)

    def _modelHasMedia(self, model, fname) -> bool:
        # First check the styling
        if fname in model["css"]:
            return True
        # If no reference to fname then check the templates as well
        for t in model["tmpls"]:
            if fname in t["qfmt"] or fname in t["afmt"]:
                return True
        return False


# Packaged Anki decks
######################################################################


class AnkiPackageExporter(AnkiExporter):
    ext = ".apkg"

    def __init__(self, col: Collection) -> None:
        AnkiExporter.__init__(self, col)

    @staticmethod
    def key(col: Collection) -> str:
        return col.tr.exporting_anki_deck_package()

    def exportInto(self, path: str) -> None:
        # open a zip file
        z = zipfile.ZipFile(
            path, "w", zipfile.ZIP_DEFLATED, allowZip64=True, strict_timestamps=False
        )
        media = self.doExport(z, path)
        # media map
        z.writestr("media", json.dumps(media))
        z.close()

    def doExport(self, z: ZipFile, path: str) -> dict[str, str]:  # type: ignore
        # export into the anki2 file
        colfile = path.replace(".apkg", ".anki2")
        AnkiExporter.exportInto(self, colfile)
        # prevent older clients from accessing

        self._addDummyCollection(z)
        z.write(colfile, "collection.anki21")

        # and media
        self.prepareMedia()
        media = self._exportMedia(z, self.mediaFiles, self.mediaDir)
        # tidy up intermediate files
        os.unlink(colfile)
        p = path.replace(".apkg", ".media.db2")
        if os.path.exists(p):
            os.unlink(p)
        shutil.rmtree(path.replace(".apkg", ".media"))
        return media

    def _exportMedia(self, z: ZipFile, files: list[str], fdir: str) -> dict[str, str]:
        media = {}
        for c, file in enumerate(files):
            cStr = str(c)
            file = hooks.media_file_filter(file)
            mpath = os.path.join(fdir, file)
            if os.path.isdir(mpath):
                continue
            if os.path.exists(mpath):
                if re.search(r"\.svg$", file, re.IGNORECASE):
                    z.write(mpath, cStr, zipfile.ZIP_DEFLATED)
                else:
                    z.write(mpath, cStr, zipfile.ZIP_STORED)
                media[cStr] = unicodedata.normalize("NFC", file)
                hooks.media_files_did_export(c)

        return media

    def prepareMedia(self) -> None:
        # chance to move each file in self.mediaFiles into place before media
        # is zipped up
        pass

    # create a dummy collection to ensure older clients don't try to read
    # data they don't understand
    def _addDummyCollection(self, zip) -> None:
        path = namedtmp("dummy.anki2")
        c = Collection(path)
        n = c.newNote()
        n.fields[0] = "This file requires a newer version of Anki."
        c.addNote(n)
        c.close(downgrade=True)

        zip.write(path, "collection.anki2")
        os.unlink(path)


# Collection package
######################################################################


class AnkiCollectionPackageExporter(AnkiPackageExporter):
    ext = ".colpkg"
    verbatim = True
    includeSched = None
    LEGACY = True

    def __init__(self, col):
        AnkiPackageExporter.__init__(self, col)

    @staticmethod
    def key(col: Collection) -> str:
        return col.tr.exporting_anki_collection_package()

    def exportInto(self, path: str) -> None:
        """Export collection. Caller must re-open afterwards."""

        def exporting_media() -> bool:
            return any(
                hook.__name__ == "exported_media"
                for hook in hooks.legacy_export_progress._hooks
            )

        def progress() -> None:
            while exporting_media():
                progress = self.col._backend.latest_progress()
                if progress.HasField("exporting"):
                    hooks.legacy_export_progress(progress.exporting)
                time.sleep(0.1)

        threading.Thread(target=progress).start()
        self.col.export_collection_package(path, self.includeMedia, self.LEGACY)


class AnkiCollectionPackage21bExporter(AnkiCollectionPackageExporter):
    LEGACY = False

    @staticmethod
    def key(_col: Collection) -> str:
        return "Anki 2.1.50+ Collection Package"


# Export modules
##########################################################################


def exporters(col: Collection) -> list[tuple[str, Any]]:
    def id(obj) -> tuple[str, Exporter]:
        if callable(obj.key):
            key_str = obj.key(col)
        else:
            key_str = obj.key
        return (f"{key_str} (*{obj.ext})", obj)

    exps = [
        id(AnkiCollectionPackageExporter),
        id(AnkiCollectionPackage21bExporter),
        id(AnkiPackageExporter),
        id(TextNoteExporter),
        id(TextCardExporter),
    ]
    hooks.exporters_list_created(exps)
    return exps



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/notetypes.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/notetypes.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import collection_pb2 as anki_dot_collection__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x61nki/notetypes.proto\x12\x0e\x61nki.notetypes\x1a\x12\x61nki/generic.proto\x1a\x15\x61nki/collection.proto\"\x1a\n\nNotetypeId\x12\x0c\n\x04ntid\x18\x01 \x01(\x03\"\xa1\x0c\n\x08Notetype\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\nmtime_secs\x18\x03 \x01(\x03\x12\x0b\n\x03usn\x18\x04 \x01(\x11\x12/\n\x06\x63onfig\x18\x07 \x01(\x0b\x32\x1f.anki.notetypes.Notetype.Config\x12.\n\x06\x66ields\x18\x08 \x03(\x0b\x32\x1e.anki.notetypes.Notetype.Field\x12\x34\n\ttemplates\x18\t \x03(\x0b\x32!.anki.notetypes.Notetype.Template\x1a\xdb\x04\n\x06\x43onfig\x12\x32\n\x04kind\x18\x01 \x01(\x0e\x32$.anki.notetypes.Notetype.Config.Kind\x12\x16\n\x0esort_field_idx\x18\x02 \x01(\r\x12\x0b\n\x03\x63ss\x18\x03 \x01(\t\x12\x1d\n\x15target_deck_id_unused\x18\x04 \x01(\x03\x12\x11\n\tlatex_pre\x18\x05 \x01(\t\x12\x12\n\nlatex_post\x18\x06 \x01(\t\x12\x11\n\tlatex_svg\x18\x07 \x01(\x08\x12=\n\x04reqs\x18\x08 \x03(\x0b\x32/.anki.notetypes.Notetype.Config.CardRequirement\x12L\n\x13original_stock_kind\x18\t \x01(\x0e\x32/.anki.notetypes.StockNotetype.OriginalStockKind\x12\x18\n\x0boriginal_id\x18\n \x01(\x03H\x00\x88\x01\x01\x12\x0e\n\x05other\x18\xff\x01 \x01(\x0c\x1a\xae\x01\n\x0f\x43\x61rdRequirement\x12\x10\n\x08\x63\x61rd_ord\x18\x01 \x01(\r\x12\x42\n\x04kind\x18\x02 \x01(\x0e\x32\x34.anki.notetypes.Notetype.Config.CardRequirement.Kind\x12\x12\n\nfield_ords\x18\x03 \x03(\r\"1\n\x04Kind\x12\r\n\tKIND_NONE\x10\x00\x12\x0c\n\x08KIND_ANY\x10\x01\x12\x0c\n\x08KIND_ALL\x10\x02\"\'\n\x04Kind\x12\x0f\n\x0bKIND_NORMAL\x10\x00\x12\x0e\n\nKIND_CLOZE\x10\x01\x42\x0e\n\x0c_original_id\x1a\xf2\x02\n\x05\x46ield\x12!\n\x03ord\x18\x01 \x01(\x0b\x32\x14.anki.generic.UInt32\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x35\n\x06\x63onfig\x18\x05 \x01(\x0b\x32%.anki.notetypes.Notetype.Field.Config\x1a\x80\x02\n\x06\x43onfig\x12\x0e\n\x06sticky\x18\x01 \x01(\x08\x12\x0b\n\x03rtl\x18\x02 \x01(\x08\x12\x11\n\tfont_name\x18\x03 \x01(\t\x12\x11\n\tfont_size\x18\x04 \x01(\r\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x12\n\nplain_text\x18\x06 \x01(\x08\x12\x11\n\tcollapsed\x18\x07 \x01(\x08\x12\x1b\n\x13\x65xclude_from_search\x18\x08 \x01(\x08\x12\x0f\n\x02id\x18\t \x01(\x03H\x00\x88\x01\x01\x12\x10\n\x03tag\x18\n \x01(\rH\x01\x88\x01\x01\x12\x18\n\x10prevent_deletion\x18\x0b \x01(\x08\x12\x0e\n\x05other\x18\xff\x01 \x01(\x0c\x42\x05\n\x03_idB\x06\n\x04_tag\x1a\xef\x02\n\x08Template\x12!\n\x03ord\x18\x01 \x01(\x0b\x32\x14.anki.generic.UInt32\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\nmtime_secs\x18\x03 \x01(\x03\x12\x0b\n\x03usn\x18\x04 \x01(\x11\x12\x38\n\x06\x63onfig\x18\x05 \x01(\x0b\x32(.anki.notetypes.Notetype.Template.Config\x1a\xd6\x01\n\x06\x43onfig\x12\x10\n\x08q_format\x18\x01 \x01(\t\x12\x10\n\x08\x61_format\x18\x02 \x01(\t\x12\x18\n\x10q_format_browser\x18\x03 \x01(\t\x12\x18\n\x10\x61_format_browser\x18\x04 \x01(\t\x12\x16\n\x0etarget_deck_id\x18\x05 \x01(\x03\x12\x19\n\x11\x62rowser_font_name\x18\x06 \x01(\t\x12\x19\n\x11\x62rowser_font_size\x18\x07 \x01(\r\x12\x0f\n\x02id\x18\x08 \x01(\x03H\x00\x88\x01\x01\x12\x0e\n\x05other\x18\xff\x01 \x01(\x0c\x42\x05\n\x03_id\"_\n\x1a\x41\x64\x64OrUpdateNotetypeRequest\x12\x0c\n\x04json\x18\x01 \x01(\x0c\x12\x1e\n\x16preserve_usn_and_mtime\x18\x02 \x01(\x08\x12\x13\n\x0bskip_checks\x18\x03 \x01(\x08\"@\n\x1bUpdateNotetypeLegacyRequest\x12\x0c\n\x04json\x18\x01 \x01(\x0c\x12\x13\n\x0bskip_checks\x18\x02 \x01(\x08\"\xfb\x03\n\rStockNotetype\x12\x30\n\x04kind\x18\x01 \x01(\x0e\x32\".anki.notetypes.StockNotetype.Kind\"\x96\x01\n\x04Kind\x12\x0e\n\nKIND_BASIC\x10\x00\x12\x1b\n\x17KIND_BASIC_AND_REVERSED\x10\x01\x12 \n\x1cKIND_BASIC_OPTIONAL_REVERSED\x10\x02\x12\x15\n\x11KIND_BASIC_TYPING\x10\x03\x12\x0e\n\nKIND_CLOZE\x10\x04\x12\x18\n\x14KIND_IMAGE_OCCLUSION\x10\x05\"\x9e\x02\n\x11OriginalStockKind\x12\x1f\n\x1bORIGINAL_STOCK_KIND_UNKNOWN\x10\x00\x12\x1d\n\x19ORIGINAL_STOCK_KIND_BASIC\x10\x01\x12*\n&ORIGINAL_STOCK_KIND_BASIC_AND_REVERSED\x10\x02\x12/\n+ORIGINAL_STOCK_KIND_BASIC_OPTIONAL_REVERSED\x10\x03\x12$\n ORIGINAL_STOCK_KIND_BASIC_TYPING\x10\x04\x12\x1d\n\x19ORIGINAL_STOCK_KIND_CLOZE\x10\x05\x12\'\n#ORIGINAL_STOCK_KIND_IMAGE_OCCLUSION\x10\x06\"@\n\rNotetypeNames\x12/\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1e.anki.notetypes.NotetypeNameId\"L\n\x11NotetypeUseCounts\x12\x37\n\x07\x65ntries\x18\x01 \x03(\x0b\x32&.anki.notetypes.NotetypeNameIdUseCount\"*\n\x0eNotetypeNameId\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\"E\n\x16NotetypeNameIdUseCount\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x11\n\tuse_count\x18\x03 \x01(\r\"1\n\x16GetAuxConfigKeyRequest\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0b\n\x03key\x18\x02 \x01(\t\"X\n\x1eGetAuxTemplateConfigKeyRequest\x12\x13\n\x0bnotetype_id\x18\x01 \x01(\x03\x12\x14\n\x0c\x63\x61rd_ordinal\x18\x02 \x01(\r\x12\x0b\n\x03key\x18\x03 \x01(\t\"P\n\x1cGetChangeNotetypeInfoRequest\x12\x17\n\x0fold_notetype_id\x18\x01 \x01(\x03\x12\x17\n\x0fnew_notetype_id\x18\x02 \x01(\x03\"\xcb\x01\n\x15\x43hangeNotetypeRequest\x12\x10\n\x08note_ids\x18\x01 \x03(\x03\x12\x12\n\nnew_fields\x18\x02 \x03(\x05\x12\x15\n\rnew_templates\x18\x03 \x03(\x05\x12\x17\n\x0fold_notetype_id\x18\x04 \x01(\x03\x12\x17\n\x0fnew_notetype_id\x18\x05 \x01(\x03\x12\x16\n\x0e\x63urrent_schema\x18\x06 \x01(\x03\x12\x19\n\x11old_notetype_name\x18\x07 \x01(\t\x12\x10\n\x08is_cloze\x18\x08 \x01(\x08\"\xcf\x01\n\x12\x43hangeNotetypeInfo\x12\x17\n\x0fold_field_names\x18\x01 \x03(\t\x12\x1a\n\x12old_template_names\x18\x02 \x03(\t\x12\x17\n\x0fnew_field_names\x18\x03 \x03(\t\x12\x1a\n\x12new_template_names\x18\x04 \x03(\t\x12\x34\n\x05input\x18\x05 \x01(\x0b\x32%.anki.notetypes.ChangeNotetypeRequest\x12\x19\n\x11old_notetype_name\x18\x06 \x01(\t\"\x9c\x01\n\x1dRestoreNotetypeToStockRequest\x12/\n\x0bnotetype_id\x18\x01 \x01(\x0b\x32\x1a.anki.notetypes.NotetypeId\x12;\n\nforce_kind\x18\x02 \x01(\x0e\x32\".anki.notetypes.StockNotetype.KindH\x00\x88\x01\x01\x42\r\n\x0b_force_kind\")\n\x19GetClozeFieldOrdsResponse\x12\x0c\n\x04ords\x18\x01 \x03(\r*\xc8\x01\n\x13ImageOcclusionField\x12$\n IMAGE_OCCLUSION_FIELD_OCCLUSIONS\x10\x00\x12\x1f\n\x1bIMAGE_OCCLUSION_FIELD_IMAGE\x10\x01\x12 \n\x1cIMAGE_OCCLUSION_FIELD_HEADER\x10\x02\x12$\n IMAGE_OCCLUSION_FIELD_BACK_EXTRA\x10\x03\x12\"\n\x1eIMAGE_OCCLUSION_FIELD_COMMENTS\x10\x04*>\n\nClozeField\x12\x14\n\x10\x43LOZE_FIELD_TEXT\x10\x00\x12\x1a\n\x16\x43LOZE_FIELD_BACK_EXTRA\x10\x01\x32\xb9\x0c\n\x10NotetypesService\x12I\n\x0b\x41\x64\x64Notetype\x12\x18.anki.notetypes.Notetype\x1a .anki.collection.OpChangesWithId\x12\x46\n\x0eUpdateNotetype\x12\x18.anki.notetypes.Notetype\x1a\x1a.anki.collection.OpChanges\x12I\n\x11\x41\x64\x64NotetypeLegacy\x12\x12.anki.generic.Json\x1a .anki.collection.OpChangesWithId\x12_\n\x14UpdateNotetypeLegacy\x12+.anki.notetypes.UpdateNotetypeLegacyRequest\x1a\x1a.anki.collection.OpChanges\x12]\n\x13\x41\x64\x64OrUpdateNotetype\x12*.anki.notetypes.AddOrUpdateNotetypeRequest\x1a\x1a.anki.notetypes.NotetypeId\x12K\n\x16GetStockNotetypeLegacy\x12\x1d.anki.notetypes.StockNotetype\x1a\x12.anki.generic.Json\x12\x43\n\x0bGetNotetype\x12\x1a.anki.notetypes.NotetypeId\x1a\x18.anki.notetypes.Notetype\x12\x43\n\x11GetNotetypeLegacy\x12\x1a.anki.notetypes.NotetypeId\x1a\x12.anki.generic.Json\x12\x46\n\x10GetNotetypeNames\x12\x13.anki.generic.Empty\x1a\x1d.anki.notetypes.NotetypeNames\x12S\n\x19GetNotetypeNamesAndCounts\x12\x13.anki.generic.Empty\x1a!.anki.notetypes.NotetypeUseCounts\x12G\n\x13GetNotetypeIdByName\x12\x14.anki.generic.String\x1a\x1a.anki.notetypes.NotetypeId\x12H\n\x0eRemoveNotetype\x12\x1a.anki.notetypes.NotetypeId\x1a\x1a.anki.collection.OpChanges\x12W\n\x17GetAuxNotetypeConfigKey\x12&.anki.notetypes.GetAuxConfigKeyRequest\x1a\x14.anki.generic.String\x12_\n\x17GetAuxTemplateConfigKey\x12..anki.notetypes.GetAuxTemplateConfigKeyRequest\x1a\x14.anki.generic.String\x12i\n\x15GetChangeNotetypeInfo\x12,.anki.notetypes.GetChangeNotetypeInfoRequest\x1a\".anki.notetypes.ChangeNotetypeInfo\x12S\n\x0e\x43hangeNotetype\x12%.anki.notetypes.ChangeNotetypeRequest\x1a\x1a.anki.collection.OpChanges\x12\x45\n\rGetFieldNames\x12\x1a.anki.notetypes.NotetypeId\x1a\x18.anki.generic.StringList\x12\x63\n\x16RestoreNotetypeToStock\x12-.anki.notetypes.RestoreNotetypeToStockRequest\x1a\x1a.anki.collection.OpChanges\x12Z\n\x11GetClozeFieldOrds\x12\x1a.anki.notetypes.NotetypeId\x1a).anki.notetypes.GetClozeFieldOrdsResponse2\x19\n\x17\x42\x61\x63kendNotetypesServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.notetypes_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_IMAGEOCCLUSIONFIELD']._serialized_start=3457
  _globals['_IMAGEOCCLUSIONFIELD']._serialized_end=3657
  _globals['_CLOZEFIELD']._serialized_start=3659
  _globals['_CLOZEFIELD']._serialized_end=3721
  _globals['_NOTETYPEID']._serialized_start=83
  _globals['_NOTETYPEID']._serialized_end=109
  _globals['_NOTETYPE']._serialized_start=112
  _globals['_NOTETYPE']._serialized_end=1681
  _globals['_NOTETYPE_CONFIG']._serialized_start=335
  _globals['_NOTETYPE_CONFIG']._serialized_end=938
  _globals['_NOTETYPE_CONFIG_CARDREQUIREMENT']._serialized_start=707
  _globals['_NOTETYPE_CONFIG_CARDREQUIREMENT']._serialized_end=881
  _globals['_NOTETYPE_CONFIG_CARDREQUIREMENT_KIND']._serialized_start=832
  _globals['_NOTETYPE_CONFIG_CARDREQUIREMENT_KIND']._serialized_end=881
  _globals['_NOTETYPE_CONFIG_KIND']._serialized_start=883
  _globals['_NOTETYPE_CONFIG_KIND']._serialized_end=922
  _globals['_NOTETYPE_FIELD']._serialized_start=941
  _globals['_NOTETYPE_FIELD']._serialized_end=1311
  _globals['_NOTETYPE_FIELD_CONFIG']._serialized_start=1055
  _globals['_NOTETYPE_FIELD_CONFIG']._serialized_end=1311
  _globals['_NOTETYPE_TEMPLATE']._serialized_start=1314
  _globals['_NOTETYPE_TEMPLATE']._serialized_end=1681
  _globals['_NOTETYPE_TEMPLATE_CONFIG']._serialized_start=1467
  _globals['_NOTETYPE_TEMPLATE_CONFIG']._serialized_end=1681
  _globals['_ADDORUPDATENOTETYPEREQUEST']._serialized_start=1683
  _globals['_ADDORUPDATENOTETYPEREQUEST']._serialized_end=1778
  _globals['_UPDATENOTETYPELEGACYREQUEST']._serialized_start=1780
  _globals['_UPDATENOTETYPELEGACYREQUEST']._serialized_end=1844
  _globals['_STOCKNOTETYPE']._serialized_start=1847
  _globals['_STOCKNOTETYPE']._serialized_end=2354
  _globals['_STOCKNOTETYPE_KIND']._serialized_start=1915
  _globals['_STOCKNOTETYPE_KIND']._serialized_end=2065
  _globals['_STOCKNOTETYPE_ORIGINALSTOCKKIND']._serialized_start=2068
  _globals['_STOCKNOTETYPE_ORIGINALSTOCKKIND']._serialized_end=2354
  _globals['_NOTETYPENAMES']._serialized_start=2356
  _globals['_NOTETYPENAMES']._serialized_end=2420
  _globals['_NOTETYPEUSECOUNTS']._serialized_start=2422
  _globals['_NOTETYPEUSECOUNTS']._serialized_end=2498
  _globals['_NOTETYPENAMEID']._serialized_start=2500
  _globals['_NOTETYPENAMEID']._serialized_end=2542
  _globals['_NOTETYPENAMEIDUSECOUNT']._serialized_start=2544
  _globals['_NOTETYPENAMEIDUSECOUNT']._serialized_end=2613
  _globals['_GETAUXCONFIGKEYREQUEST']._serialized_start=2615
  _globals['_GETAUXCONFIGKEYREQUEST']._serialized_end=2664
  _globals['_GETAUXTEMPLATECONFIGKEYREQUEST']._serialized_start=2666
  _globals['_GETAUXTEMPLATECONFIGKEYREQUEST']._serialized_end=2754
  _globals['_GETCHANGENOTETYPEINFOREQUEST']._serialized_start=2756
  _globals['_GETCHANGENOTETYPEINFOREQUEST']._serialized_end=2836
  _globals['_CHANGENOTETYPEREQUEST']._serialized_start=2839
  _globals['_CHANGENOTETYPEREQUEST']._serialized_end=3042
  _globals['_CHANGENOTETYPEINFO']._serialized_start=3045
  _globals['_CHANGENOTETYPEINFO']._serialized_end=3252
  _globals['_RESTORENOTETYPETOSTOCKREQUEST']._serialized_start=3255
  _globals['_RESTORENOTETYPETOSTOCKREQUEST']._serialized_end=3411
  _globals['_GETCLOZEFIELDORDSRESPONSE']._serialized_start=3413
  _globals['_GETCLOZEFIELDORDSRESPONSE']._serialized_end=3454
  _globals['_NOTETYPESSERVICE']._serialized_start=3724
  _globals['_NOTETYPESSERVICE']._serialized_end=5317
  _globals['_BACKENDNOTETYPESSERVICE']._serialized_start=5319
  _globals['_BACKENDNOTETYPESSERVICE']._serialized_end=5344
# @@protoc_insertion_point(module_scope)



# Form implementation generated from reading ui file 'qt/aqt/forms/exporting.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_ExportDialog(object):
    def setupUi(self, ExportDialog):
        ExportDialog.setObjectName("ExportDialog")
        ExportDialog.resize(550, 200)
        self.vboxlayout = QtWidgets.QVBoxLayout(ExportDialog)
        self.vboxlayout.setObjectName("vboxlayout")
        self.gridlayout = QtWidgets.QGridLayout()
        self.gridlayout.setObjectName("gridlayout")
        self.label = QtWidgets.QLabel(parent=ExportDialog)
        self.label.setMinimumSize(QtCore.QSize(100, 0))
        self.label.setObjectName("label")
        self.gridlayout.addWidget(self.label, 0, 0, 1, 1)
        self.format = QtWidgets.QComboBox(parent=ExportDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.format.sizePolicy().hasHeightForWidth())
        self.format.setSizePolicy(sizePolicy)
        self.format.setObjectName("format")
        self.gridlayout.addWidget(self.format, 0, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(parent=ExportDialog)
        self.label_2.setObjectName("label_2")
        self.gridlayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.deck = QtWidgets.QComboBox(parent=ExportDialog)
        self.deck.setMinimumContentsLength(50)
        self.deck.setObjectName("deck")
        self.gridlayout.addWidget(self.deck, 1, 1, 1, 1)
        self.vboxlayout.addLayout(self.gridlayout)
        self.vboxlayout1 = QtWidgets.QVBoxLayout()
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.includeSched = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeSched.setChecked(True)
        self.includeSched.setObjectName("includeSched")
        self.vboxlayout1.addWidget(self.includeSched, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.include_deck_configs = QtWidgets.QCheckBox(parent=ExportDialog)
        self.include_deck_configs.setChecked(False)
        self.include_deck_configs.setObjectName("include_deck_configs")
        self.vboxlayout1.addWidget(self.include_deck_configs, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeMedia = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeMedia.setChecked(True)
        self.includeMedia.setObjectName("includeMedia")
        self.vboxlayout1.addWidget(self.includeMedia, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeHTML = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeHTML.setObjectName("includeHTML")
        self.vboxlayout1.addWidget(self.includeHTML, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeTags = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeTags.setChecked(True)
        self.includeTags.setObjectName("includeTags")
        self.vboxlayout1.addWidget(self.includeTags, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeDeck = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeDeck.setEnabled(True)
        self.includeDeck.setObjectName("includeDeck")
        self.vboxlayout1.addWidget(self.includeDeck, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeNotetype = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeNotetype.setEnabled(True)
        self.includeNotetype.setObjectName("includeNotetype")
        self.vboxlayout1.addWidget(self.includeNotetype, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.includeGuid = QtWidgets.QCheckBox(parent=ExportDialog)
        self.includeGuid.setObjectName("includeGuid")
        self.vboxlayout1.addWidget(self.includeGuid, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.legacy_support = QtWidgets.QCheckBox(parent=ExportDialog)
        self.legacy_support.setChecked(False)
        self.legacy_support.setObjectName("legacy_support")
        self.vboxlayout1.addWidget(self.legacy_support, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.vboxlayout.addLayout(self.vboxlayout1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.vboxlayout.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=ExportDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)

        self.retranslateUi(ExportDialog)
        self.buttonBox.accepted.connect(ExportDialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(ExportDialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(ExportDialog)
        ExportDialog.setTabOrder(self.format, self.deck)
        ExportDialog.setTabOrder(self.deck, self.includeSched)
        ExportDialog.setTabOrder(self.includeSched, self.include_deck_configs)
        ExportDialog.setTabOrder(self.include_deck_configs, self.includeMedia)
        ExportDialog.setTabOrder(self.includeMedia, self.includeHTML)
        ExportDialog.setTabOrder(self.includeHTML, self.includeTags)
        ExportDialog.setTabOrder(self.includeTags, self.includeDeck)
        ExportDialog.setTabOrder(self.includeDeck, self.includeNotetype)
        ExportDialog.setTabOrder(self.includeNotetype, self.includeGuid)
        ExportDialog.setTabOrder(self.includeGuid, self.legacy_support)

    def retranslateUi(self, ExportDialog):
        _translate = QtCore.QCoreApplication.translate
        ExportDialog.setWindowTitle(tr.actions_export())
        self.label.setText(tr.exporting_export_format())
        self.label_2.setText(tr.exporting_include())
        self.includeSched.setText(tr.exporting_include_scheduling_information())
        self.include_deck_configs.setText(tr.exporting_include_deck_configs())
        self.includeMedia.setText(tr.exporting_include_media())
        self.includeHTML.setText(tr.exporting_include_html_and_media_references())
        self.includeTags.setText(tr.exporting_include_tags())
        self.includeDeck.setText(tr.exporting_include_deck())
        self.includeNotetype.setText(tr.exporting_include_notetype())
        self.includeGuid.setText(tr.exporting_include_guid())
        self.legacy_support.setText(tr.exporting_support_older_anki_versions())


# Form implementation generated from reading ui file 'qt/aqt/forms/clayout_top.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(12)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.card_type_label = QtWidgets.QLabel(parent=Form)
        self.card_type_label.setText("CARD TYPE:")
        self.card_type_label.setObjectName("card_type_label")
        self.horizontalLayout.addWidget(self.card_type_label)
        self.templatesBox = QtWidgets.QComboBox(parent=Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(30)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.templatesBox.sizePolicy().hasHeightForWidth())
        self.templatesBox.setSizePolicy(sizePolicy)
        self.templatesBox.setMinimumContentsLength(50)
        self.templatesBox.setMaxVisibleItems(30)
        self.templatesBox.setObjectName("templatesBox")
        self.horizontalLayout.addWidget(self.templatesBox)
        spacerItem = QtWidgets.QSpacerItem(1, 20, QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.templateOptions = QtWidgets.QPushButton(parent=Form)
        self.templateOptions.setText("")
        self.templateOptions.setAutoDefault(False)
        self.templateOptions.setDefault(False)
        self.templateOptions.setObjectName("templateOptions")
        self.horizontalLayout.addWidget(self.templateOptions)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(tr.card_templates_form())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import anki.collection
import anki.models
from anki import notetypes_pb2
from anki._legacy import DeprecatedNamesMixinForModule
from anki.utils import from_json_bytes

StockNotetypeKind = notetypes_pb2.StockNotetype.Kind

# add-on authors can add ("note type name", function)
# to this list to have it shown in the add/clone note type screen
models: list[tuple] = []


def _get_stock_notetype(
    col: anki.collection.Collection, kind: StockNotetypeKind.V
) -> anki.models.NotetypeDict:
    return from_json_bytes(col._backend.get_stock_notetype_legacy(kind))


def get_stock_notetypes(
    col: anki.collection.Collection,
) -> list[tuple[str, Callable[[anki.collection.Collection], anki.models.NotetypeDict]]]:
    out: list[
        tuple[str, Callable[[anki.collection.Collection], anki.models.NotetypeDict]]
    ] = []
    # add standard - this order should match the one in notetypes.proto
    for kind in [
        StockNotetypeKind.KIND_BASIC,
        StockNotetypeKind.KIND_BASIC_AND_REVERSED,
        StockNotetypeKind.KIND_BASIC_OPTIONAL_REVERSED,
        StockNotetypeKind.KIND_BASIC_TYPING,
        StockNotetypeKind.KIND_CLOZE,
        StockNotetypeKind.KIND_IMAGE_OCCLUSION,
    ]:
        note_type = from_json_bytes(col._backend.get_stock_notetype_legacy(kind))

        def instance_getter(
            model: Any,
        ) -> Callable[[anki.collection.Collection], anki.models.NotetypeDict]:
            return lambda col: model

        out.append((note_type["name"], instance_getter(note_type)))
    # add extras from add-ons
    for name_or_func, func in models:
        if not isinstance(name_or_func, str):
            name = name_or_func()
        else:
            name = name_or_func
        out.append((name, func))
    return out


#
# Legacy functions that added the notetype before returning it
#


def _legacy_add_basic_model(
    col: anki.collection.Collection,
) -> anki.models.NotetypeDict:
    note_type = _get_stock_notetype(col, StockNotetypeKind.KIND_BASIC)
    col.models.add(note_type)
    return note_type


def _legacy_add_basic_typing_model(
    col: anki.collection.Collection,
) -> anki.models.NotetypeDict:
    note_type = _get_stock_notetype(col, StockNotetypeKind.KIND_BASIC_TYPING)
    col.models.add(note_type)
    return note_type


def _legacy_add_forward_reverse(
    col: anki.collection.Collection,
) -> anki.models.NotetypeDict:
    note_type = _get_stock_notetype(col, StockNotetypeKind.KIND_BASIC_AND_REVERSED)
    col.models.add(note_type)
    return note_type


def _legacy_add_forward_optional_reverse(
    col: anki.collection.Collection,
) -> anki.models.NotetypeDict:
    note_type = _get_stock_notetype(col, StockNotetypeKind.KIND_BASIC_OPTIONAL_REVERSED)
    col.models.add(note_type)
    return note_type


def _legacy_add_cloze_model(
    col: anki.collection.Collection,
) -> anki.models.NotetypeDict:
    note_type = _get_stock_notetype(col, StockNotetypeKind.KIND_CLOZE)
    col.models.add(note_type)
    return note_type


_deprecated_names = DeprecatedNamesMixinForModule(globals())
_deprecated_names.register_deprecated_attributes(
    addBasicModel=(_legacy_add_basic_model, get_stock_notetypes),
    addBasicTypingModel=(_legacy_add_basic_typing_model, get_stock_notetypes),
    addForwardReverse=(_legacy_add_forward_reverse, get_stock_notetypes),
    addForwardOptionalReverse=(
        _legacy_add_forward_optional_reverse,
        get_stock_notetypes,
    ),
    addClozeModel=(_legacy_add_cloze_model, get_stock_notetypes),
)


if not TYPE_CHECKING:

    def __getattr__(name: str) -> Any:
        return _deprecated_names.__getattr__(name)



# Form implementation generated from reading ui file 'qt/aqt/forms/addonconf.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        Dialog.resize(631, 521)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QtWidgets.QSplitter(parent=Dialog)
        self.splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.splitter.setObjectName("splitter")
        self.editor = QtWidgets.QPlainTextEdit(parent=self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.editor.sizePolicy().hasHeightForWidth())
        self.editor.setSizePolicy(sizePolicy)
        self.editor.setLineWrapMode(QtWidgets.QPlainTextEdit.LineWrapMode.NoWrap)
        self.editor.setObjectName("editor")
        self.help = AnkiWebView(parent=self.splitter)
        self.help.setProperty("url", QtCore.QUrl("about:blank"))
        self.help.setObjectName("help")
        self.verticalLayout.addWidget(self.splitter)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok|QtWidgets.QDialogButtonBox.StandardButton.RestoreDefaults)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.addons_configuration())
from aqt.webview import AnkiWebView


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import json
from collections.abc import Callable

from google.protobuf.json_format import MessageToDict

import aqt
from anki.cards import Card, CardId
from anki.errors import NotFoundError
from anki.lang import without_unicode_isolation
from aqt.qt import *
from aqt.utils import (
    add_close_shortcut,
    disable_help_button,
    qconnect,
    restoreGeom,
    saveGeom,
    setWindowIcon,
    tooltip,
    tr,
)
from aqt.webview import AnkiWebView, AnkiWebViewKind


class CardInfoDialog(QDialog):
    TITLE = "browser card info"
    GEOMETRY_KEY = "revlog"
    silentlyClose = True

    def __init__(
        self,
        parent: QWidget | None,
        mw: aqt.AnkiQt,
        card: Card | None,
        on_close: Callable | None = None,
        geometry_key: str | None = None,
        window_title: str | None = None,
    ) -> None:
        super().__init__(parent)
        self.mw = mw
        self._on_close = on_close
        self.GEOMETRY_KEY = geometry_key or self.GEOMETRY_KEY
        if window_title:
            self.setWindowTitle(window_title)
        self._setup_ui(card.id if card else None)
        self.show()

    def _setup_ui(self, card_id: CardId | None) -> None:
        self.mw.garbage_collect_on_dialog_finish(self)
        self.setMinimumSize(400, 300)
        disable_help_button(self)
        restoreGeom(self, self.GEOMETRY_KEY, default_size=(800, 800))
        add_close_shortcut(self)
        setWindowIcon(self)

        self.web: AnkiWebView | None = AnkiWebView(
            kind=AnkiWebViewKind.BROWSER_CARD_INFO
        )
        self.web.setVisible(False)
        self.web.load_sveltekit_page(f"card-info/{card_id}")
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.web)
        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        buttons.setContentsMargins(10, 0, 10, 10)
        layout.addWidget(buttons)
        qconnect(buttons.rejected, self.reject)

        self.copy_debug_info = QShortcut(  # type: ignore
            "ctrl+c", self, activated=lambda: self.copy_card_info(card_id)
        )

        self.setLayout(layout)

    def copy_card_info(self, card_id: CardId | None) -> None:
        if self.web and self.web.selectedText():
            self.web.onCopy()
            return
        if card_id is None:
            return
        assert aqt.mw.col.db, tr.errors_inconsistent_db_state()

        proto_info = aqt.mw.col.card_stats_data(card_id)
        info = MessageToDict(proto_info)

        card = aqt.mw.col.get_card(card_id)

        revlog = aqt.mw.col.db.execute(
            f"SELECT * FROM revlog WHERE cid == {card_id} ORDER BY id DESC"
        )
        deck = aqt.mw.col.decks.get(card.did) or dict()
        config = aqt.mw.col.decks.get_config(deck.get("conf", -1)) or dict()

        info["deck"] = deck
        info["config"] = config

        info["config"].pop("name", None)
        info["deck"].pop("name", None)
        info["deck"].pop("desc", None)
        info["deck"].pop("usn", None)
        info.pop("usn", None)
        info.pop("cardType", None)
        info.pop("notetype", None)
        info.pop("preset", None)

        info["cardRow"] = aqt.mw.col.db.execute(
            f"SELECT * FROM cards WHERE id == {card_id} ORDER BY id DESC"
        )[0]

        new_revlog = [
            {"row": revlog, "info": card_info_review}
            for revlog, card_info_review in zip(revlog, info.get("revlog", []))
        ]
        info["revlog"] = new_revlog
        info["rollover"] = aqt.mw.col.get_config("rollover")

        clipboard = QApplication.clipboard()
        assert clipboard is not None
        clipboard.setText(json.dumps(info, indent=2))

        tooltip(tr.about_copied_to_clipboard())

    def update_card(self, card_id: CardId | None) -> None:
        try:
            self.mw.col.get_card(card_id)
        except NotFoundError:
            card_id = None

        assert self.web is not None
        self.web.eval(f"anki.updateCard('{card_id}');")

    def reject(self) -> None:
        if self._on_close:
            self._on_close()
        assert self.web is not None
        self.web.cleanup()
        self.web = None
        saveGeom(self, self.GEOMETRY_KEY)
        return QDialog.reject(self)


class CardInfoManager:
    """Wrapper class to conveniently toggle, update and close a card info dialog."""

    def __init__(self, mw: aqt.AnkiQt, geometry_key: str, window_title: str):
        self.mw = mw
        self.geometry_key = geometry_key
        self.window_title = window_title
        self._card: Card | None = None
        self._dialog: CardInfoDialog | None = None

    def show(self) -> None:
        if self._dialog:
            self._dialog.activateWindow()
            self._dialog.raise_()
        else:
            self._dialog = CardInfoDialog(
                None,
                self.mw,
                self._card,
                self._on_close,
                self.geometry_key,
                self.window_title,
            )

    def set_card(self, card: Card | None) -> None:
        self._card = card
        if self._dialog:
            self._dialog.update_card(card.id if card else None)

    def close(self) -> None:
        if self._dialog:
            self._dialog.reject()

    def _on_close(self) -> None:
        self._dialog = None


class BrowserCardInfo(CardInfoManager):
    def __init__(self, mw: aqt.AnkiQt):
        super().__init__(
            mw,
            "revlog",
            without_unicode_isolation(
                tr.card_stats_current_card(context=tr.qt_misc_browse())
            ),
        )


class ReviewerCardInfo(CardInfoManager):
    def __init__(self, mw: aqt.AnkiQt):
        super().__init__(
            mw,
            "reviewerCardInfo",
            without_unicode_isolation(
                tr.card_stats_current_card(context=tr.decks_study())
            ),
        )


class PreviousReviewerCardInfo(CardInfoManager):
    def __init__(self, mw: aqt.AnkiQt):
        super().__init__(
            mw,
            "previousReviewerCardInfo",
            without_unicode_isolation(
                tr.card_stats_previous_card(context=tr.decks_study())
            ),
        )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
Legacy support
"""

from __future__ import annotations

from typing import Any

import anki
import anki.sound
import anki.utils
import aqt
from aqt.theme import theme_manager

# Routines removed from pylib/
##########################################################################


def bodyClass(col, card) -> str:  # type: ignore
    print("bodyClass() deprecated")
    return theme_manager.body_classes_for_card_ord(card.ord)


def allSounds(text) -> list:  # type: ignore
    print("allSounds() deprecated")
    return aqt.mw.col.media._extract_filenames(text)


def stripSounds(text) -> str:  # type: ignore
    print("stripSounds() deprecated")
    return aqt.mw.col.media.strip_av_tags(text)


def fmtTimeSpan(
    time: Any,
    pad: Any = 0,
    point: Any = 0,
    short: Any = False,
    inTime: Any = False,
    unit: Any = 99,
) -> Any:
    print("fmtTimeSpan() has become col.format_timespan()")
    return aqt.mw.col.format_timespan(time)


def install_pylib_legacy() -> None:
    anki.utils.bodyClass = bodyClass  # type: ignore
    anki.utils.fmtTimeSpan = fmtTimeSpan  # type: ignore
    anki.sound._soundReg = r"\[sound:(.+?)\]"  # type: ignore
    anki.sound.allSounds = allSounds  # type: ignore
    anki.sound.stripSounds = stripSounds  # type: ignore



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import html
import io
import json
import logging
import os
import re
import sys
import traceback
import zipfile
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from concurrent.futures import Future
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import IO, Any, Union
from urllib.parse import parse_qs, urlparse
from zipfile import ZipFile

import jsonschema
import markdown
from jsonschema.exceptions import ValidationError
from markdown.extensions import md_in_html

import anki
import anki.utils
import aqt
import aqt.forms
import aqt.main
from anki.collection import AddonInfo
from anki.httpclient import HttpClient
from anki.lang import without_unicode_isolation
from anki.utils import int_version_to_str
from aqt import gui_hooks
from aqt.log import ADDON_LOGGER_PREFIX, find_addon_logger, get_addon_logs_folder
from aqt.qt import *
from aqt.utils import (
    add_close_shortcut,
    askUser,
    disable_help_button,
    getFile,
    openFolder,
    openLink,
    restoreGeom,
    restoreSplitter,
    saveGeom,
    saveSplitter,
    send_to_trash,
    show_info,
    showInfo,
    showText,
    showWarning,
    supportText,
    tooltip,
    tr,
)


class AbortAddonImport(Exception):
    """If raised during add-on import, Anki will silently ignore this exception.
    This allows you to terminate loading without an error being shown."""


@dataclass
class InstallOk:
    name: str
    conflicts: set[str]
    compatible: bool


@dataclass
class InstallError:
    errmsg: str


@dataclass
class DownloadOk:
    data: bytes
    filename: str
    mod_time: int
    min_point_version: int
    max_point_version: int
    branch_index: int


@dataclass
class DownloadError:
    # set if result was not 200
    status_code: int | None = None
    # set if an exception occurred
    exception: Exception | None = None


# first arg is add-on id
DownloadLogEntry = tuple[int, Union[DownloadError, InstallError, InstallOk]]


ANKIWEB_ID_RE = re.compile(r"^\d+$")

_current_version = anki.utils.int_version()


@dataclass
class AddonMeta:
    dir_name: str
    provided_name: str | None
    enabled: bool
    installed_at: int
    conflicts: list[str]
    min_version: int
    max_version: int
    branch_index: int
    human_version: str | None
    update_enabled: bool
    homepage: str | None

    def human_name(self) -> str:
        return self.provided_name or self.dir_name

    def ankiweb_id(self) -> int | None:
        m = ANKIWEB_ID_RE.match(self.dir_name)
        if m:
            return int(m.group(0))
        else:
            return None

    def compatible(self) -> bool:
        min = self.min_version
        if min is not None and _current_version < min:
            return False
        max = self.max_version
        if max is not None and max < 0 and _current_version > abs(max):
            return False
        return True

    def is_latest(self, server_update_time: int) -> bool:
        return self.installed_at >= server_update_time

    def page(self) -> str | None:
        if self.ankiweb_id():
            return f"{aqt.appShared}info/{self.dir_name}"
        return self.homepage

    @staticmethod
    def from_json_meta(dir_name: str, json_meta: dict[str, Any]) -> AddonMeta:
        return AddonMeta(
            dir_name=dir_name,
            provided_name=json_meta.get("name"),
            enabled=not json_meta.get("disabled"),
            installed_at=json_meta.get("mod", 0),
            conflicts=json_meta.get("conflicts", []),
            min_version=json_meta.get("min_point_version", 0) or 0,
            max_version=json_meta.get("max_point_version", 0) or 0,
            branch_index=json_meta.get("branch_index", 0) or 0,
            human_version=json_meta.get("human_version"),
            update_enabled=json_meta.get("update_enabled", True),
            homepage=json_meta.get("homepage"),
        )


def package_name_valid(name: str) -> bool:
    # embedded /?
    base = os.path.basename(name)
    if base != name:
        return False
    # tries to escape to parent?
    root = os.getcwd()
    subfolder = os.path.abspath(os.path.join(root, name))
    if root.startswith(subfolder):
        return False
    return True


# fixme: this class should not have any GUI code in it
class AddonManager:
    exts: list[str] = [".ankiaddon", ".zip"]
    _manifest_schema: dict = {
        "type": "object",
        "properties": {
            # the name of the folder
            "package": {"type": "string", "minLength": 1, "meta": False},
            # the displayed name to the user
            "name": {"type": "string", "meta": True},
            # the time the add-on was last modified
            "mod": {"type": "number", "meta": True},
            # a list of other packages that conflict
            "conflicts": {"type": "array", "items": {"type": "string"}, "meta": True},
            # x for anki 2.1.x; int_version() for more recent releases
            "min_point_version": {"type": "number", "meta": True},
            # x for anki 2.1.x; int_version() for more recent releases
            # if negative, abs(n) is the maximum version this add-on supports
            # if positive, indicates version tested on, and is ignored
            "max_point_version": {"type": "number", "meta": True},
            # AnkiWeb sends this to indicate which branch the user downloaded.
            "branch_index": {"type": "number", "meta": True},
            # version string set by the add-on creator
            "human_version": {"type": "string", "meta": True},
            # add-on page on AnkiWeb or some other webpage
            "homepage": {"type": "string", "meta": True},
        },
        "required": ["package", "name"],
    }

    def __init__(self, mw: aqt.main.AnkiQt) -> None:
        self.mw = mw
        self.dirty = False
        f = self.mw.form
        qconnect(f.actionAdd_ons.triggered, self.onAddonsDialog)
        sys.path.insert(0, self.addonsFolder())

    # in new code, you may want all_addon_meta() instead
    def allAddons(self) -> list[str]:
        l = []
        for d in os.listdir(self.addonsFolder()):
            path = self.addonsFolder(d)
            if not os.path.exists(os.path.join(path, "__init__.py")):
                continue
            l.append(d)
        l.sort()
        if os.getenv("ANKIREVADDONS", ""):
            l = list(reversed(l))
        return l

    def all_addon_meta(self) -> Iterable[AddonMeta]:
        return map(self.addon_meta, self.allAddons())

    def addonsFolder(self, module: str | None = None) -> str:
        root = self.mw.pm.addonFolder()
        if module is None:
            return root
        return os.path.join(root, module)

    def loadAddons(self) -> None:
        from aqt import mw

        broken: list[str] = []
        error_text = ""
        for addon in self.all_addon_meta():
            if not addon.enabled:
                continue
            if not addon.compatible():
                continue
            self.dirty = True
            try:
                __import__(addon.dir_name)
            except AbortAddonImport:
                pass
            except Exception:
                name = html.escape(addon.human_name())
                page = addon.page()
                if page:
                    broken.append(f"<a href={page}>{name}</a>")
                else:
                    broken.append(name)
                tb = traceback.format_exc()
                print(tb)
                error_text += f"When loading {name}:\n{tb}\n"

        if broken:
            addons = "\n\n- " + "\n- ".join(broken)
            error = tr.addons_failed_to_load2(
                addons=addons,
            )
            txt = f"# {tr.addons_startup_failed()}\n{error}"
            html2 = markdown.markdown(txt)
            box: QDialogButtonBox
            (diag, box) = showText(
                html2,
                type="html",
                run=False,
            )

            def on_check() -> None:
                tooltip(tr.addons_checking())

                def on_done(log: list[DownloadLogEntry]) -> None:
                    if not log:
                        tooltip(tr.addons_no_updates_available())

                mw.check_for_addon_updates(by_user=True, on_done=on_done)

            def on_copy() -> None:
                txt = supportText() + "\n" + error_text
                QApplication.clipboard().setText(txt)
                tooltip(tr.about_copied_to_clipboard(), parent=diag)

            check = box.addButton(
                tr.addons_check_for_updates(), QDialogButtonBox.ButtonRole.ActionRole
            )
            check.clicked.connect(on_check)

            copy = box.addButton(
                tr.about_copy_debug_info(), QDialogButtonBox.ButtonRole.ActionRole
            )
            copy.clicked.connect(on_copy)

            # calling show immediately appears to crash
            mw.progress.single_shot(1000, diag.show)

    def onAddonsDialog(self) -> None:
        aqt.dialogs.open("AddonsDialog", self)

    # Metadata
    ######################################################################

    def addon_meta(self, dir_name: str) -> AddonMeta:
        """Get info about an installed add-on."""
        json_obj = self.addonMeta(dir_name)
        return AddonMeta.from_json_meta(dir_name, json_obj)

    def write_addon_meta(self, addon: AddonMeta) -> None:
        # preserve any unknown attributes
        json_obj = self.addonMeta(addon.dir_name)

        if addon.provided_name is not None:
            json_obj["name"] = addon.provided_name
        json_obj["disabled"] = not addon.enabled
        json_obj["mod"] = addon.installed_at
        json_obj["conflicts"] = addon.conflicts
        json_obj["max_point_version"] = addon.max_version
        json_obj["min_point_version"] = addon.min_version
        json_obj["branch_index"] = addon.branch_index
        if addon.human_version is not None:
            json_obj["human_version"] = addon.human_version
        json_obj["update_enabled"] = addon.update_enabled

        self.writeAddonMeta(addon.dir_name, json_obj)

    def _addonMetaPath(self, module: str) -> str:
        return os.path.join(self.addonsFolder(module), "meta.json")

    # in new code, use self.addon_meta() instead
    def addonMeta(self, module: str) -> dict[str, Any]:
        path = self._addonMetaPath(module)
        try:
            with open(path, encoding="utf8") as f:
                return json.load(f)
        except json.JSONDecodeError as e:
            print(f"json error in add-on {module}:\n{e}")
            return dict()
        except Exception:
            # missing meta file, etc
            return dict()

    # in new code, use write_addon_meta() instead
    def writeAddonMeta(self, module: str, meta: dict[str, Any]) -> None:
        path = self._addonMetaPath(module)
        with open(path, "w", encoding="utf8") as f:
            json.dump(meta, f)

    def toggleEnabled(self, module: str, enable: bool | None = None) -> None:
        addon = self.addon_meta(module)
        should_enable = enable if enable is not None else not addon.enabled
        if should_enable is True:
            conflicting = self._disableConflicting(module)
            if conflicting:
                addons = ", ".join(self.addonName(f) for f in conflicting)
                showInfo(
                    tr.addons_the_following_addons_are_incompatible_with(
                        name=addon.human_name(),
                        found=addons,
                    ),
                    textFormat="plain",
                )

        addon.enabled = should_enable
        self.write_addon_meta(addon)

    def ankiweb_addons(self) -> list[int]:
        ids = []
        for meta in self.all_addon_meta():
            if meta.ankiweb_id() is not None:
                ids.append(meta.ankiweb_id())
        return ids

    # Legacy helpers
    ######################################################################

    def isEnabled(self, module: str) -> bool:
        return self.addon_meta(module).enabled

    def addonName(self, module: str) -> str:
        return self.addon_meta(module).human_name()

    def addonConflicts(self, module: str) -> list[str]:
        return self.addon_meta(module).conflicts

    def annotatedName(self, module: str) -> str:
        meta = self.addon_meta(module)
        name = meta.human_name()
        if not meta.enabled:
            name += f" {tr.addons_disabled()}"
        return name

    # Conflict resolution
    ######################################################################

    def allAddonConflicts(self) -> dict[str, list[str]]:
        all_conflicts: dict[str, list[str]] = defaultdict(list)
        for addon in self.all_addon_meta():
            if not addon.enabled:
                continue
            for other_dir in addon.conflicts:
                all_conflicts[other_dir].append(addon.dir_name)
        return all_conflicts

    def _disableConflicting(
        self, module: str, conflicts: list[str] | None = None
    ) -> set[str]:
        if not self.isEnabled(module):
            # disabled add-ons should not trigger conflict handling
            return set()

        conflicts = conflicts or self.addonConflicts(module)

        installed = self.allAddons()
        found = {d for d in conflicts if d in installed and self.isEnabled(d)}
        found.update(self.allAddonConflicts().get(module, []))

        for package in found:
            self.toggleEnabled(package, enable=False)

        return found

    # Installing and deleting add-ons
    ######################################################################

    def readManifestFile(self, zfile: ZipFile) -> dict[Any, Any]:
        try:
            with zfile.open("manifest.json") as f:
                data = json.loads(f.read())
            jsonschema.validate(data, self._manifest_schema)
            # build new manifest from recognized keys
            schema = self._manifest_schema["properties"]
            manifest = {key: data[key] for key in data.keys() & schema.keys()}
        except (KeyError, json.decoder.JSONDecodeError, ValidationError):
            # raised for missing manifest, invalid json, missing/invalid keys
            return {}
        return manifest

    def install(
        self,
        file: IO | str,
        manifest: dict[str, Any] | None = None,
        force_enable: bool = False,
    ) -> InstallOk | InstallError:
        """Install add-on from path or file-like object. Metadata is read
        from the manifest file, with keys overridden by supplying a 'manifest'
        dictionary"""
        try:
            zfile = ZipFile(file)
        except zipfile.BadZipfile:
            return InstallError(errmsg="zip")

        with zfile:
            file_manifest = self.readManifestFile(zfile)
            if manifest:
                file_manifest.update(manifest)
            manifest = file_manifest
            if not manifest:
                return InstallError(errmsg="manifest")
            package = manifest["package"]
            if not package_name_valid(package):
                return InstallError(errmsg="invalid package")
            conflicts = manifest.get("conflicts", [])
            found_conflicts = self._disableConflicting(package, conflicts)
            meta = self.addonMeta(package)
            gui_hooks.addon_manager_will_install_addon(self, package)
            self._install(package, zfile)
            gui_hooks.addon_manager_did_install_addon(self, package)

        schema = self._manifest_schema["properties"]
        manifest_meta = {
            k: v for k, v in manifest.items() if k in schema and schema[k]["meta"]
        }
        meta.update(manifest_meta)

        if force_enable:
            meta["disabled"] = False

        self.writeAddonMeta(package, meta)

        meta2 = self.addon_meta(package)

        return InstallOk(
            name=meta["name"], conflicts=found_conflicts, compatible=meta2.compatible()
        )

    def _install(self, module: str, zfile: ZipFile) -> None:
        # previously installed?
        base = self.addonsFolder(module)
        if os.path.exists(base):
            self.backupUserFiles(module)
            try:
                self.deleteAddon(module)
            except Exception:
                self.restoreUserFiles(module)
                raise
        os.mkdir(base)
        self.restoreUserFiles(module)

        # extract
        for n in zfile.namelist():
            if n.endswith("/"):
                # folder; ignore
                continue

            path = os.path.join(base, n)
            # skip existing user files
            if os.path.exists(path) and n.startswith("user_files/"):
                continue
            zfile.extract(n, base)

    def deleteAddon(self, module: str) -> None:
        send_to_trash(Path(self.addonsFolder(module)))

    # Processing local add-on files
    ######################################################################

    def processPackages(
        self,
        paths: list[str],
        parent: QWidget | None = None,
        force_enable: bool = False,
    ) -> tuple[list[str], list[str]]:
        log = []
        errs = []

        self.mw.progress.start(parent=parent)
        try:
            for path in paths:
                base = os.path.basename(path)
                result = self.install(path, force_enable=force_enable)

                if isinstance(result, InstallError):
                    errs.extend(
                        self._installationErrorReport(result, base, mode="local")
                    )
                else:
                    log.extend(
                        self._installationSuccessReport(result, base, mode="local")
                    )
        finally:
            self.mw.progress.finish()

        return log, errs

    # Installation messaging
    ######################################################################

    def _installationErrorReport(
        self, result: InstallError, base: str, mode: str = "download"
    ) -> list[str]:
        messages = {
            "zip": tr.addons_corrupt_addon_file(),
            "manifest": tr.addons_invalid_addon_manifest(),
        }

        msg = messages.get(result.errmsg, tr.addons_unknown_error(val=result.errmsg))

        if mode == "download":
            template = tr.addons_error_downloading_ids_errors(id=base, error=msg)
        else:
            template = tr.addons_error_installing_bases_errors(base=base, error=msg)

        return [template]

    def _installationSuccessReport(
        self, result: InstallOk, base: str, mode: str = "download"
    ) -> list[str]:
        name = result.name or base
        if mode == "download":
            template = tr.addons_downloaded_fnames(fname=name)
        else:
            template = tr.addons_installed_names(name=name)

        strings = [template]

        if result.conflicts:
            strings.append(
                tr.addons_the_following_conflicting_addons_were_disabled()
                + " "
                + ", ".join(self.addonName(f) for f in result.conflicts)
            )

        if not result.compatible:
            strings.append(tr.addons_this_addon_is_not_compatible_with())

        return strings

    # Updating
    ######################################################################

    def update_supported_versions(self, items: list[AddonInfo]) -> None:
        """Adjust the supported version range after an update check.

        AnkiWeb will not have sent us any add-ons that don't support our
        version, so this cannot disable add-ons that users are using. It
        does allow the add-on author to mark an add-on as not supporting
        a future release, causing the add-on to be disabled when the user
        upgrades.
        """

        for item in items:
            addon = self.addon_meta(str(item.id))
            updated = False

            if addon.max_version != item.max_version:
                addon.max_version = item.max_version
                updated = True
            if addon.min_version != item.min_version:
                addon.min_version = item.min_version
                updated = True

            if updated:
                self.write_addon_meta(addon)

    def get_updated_addons(self, items: list[AddonInfo]) -> list[AddonInfo]:
        """Return ids of add-ons requiring an update."""
        need_update = []
        for item in items:
            addon = self.addon_meta(str(item.id))
            # update if server mtime is newer
            if not addon.is_latest(item.modified):
                need_update.append(item)
            elif not addon.compatible():
                # Addon is currently disabled, and a suitable branch was found on the
                # server. Ignore our stored mtime (which may have been set incorrectly
                # in the past) and require an update.
                need_update.append(item)

        return need_update

    # Add-on Config
    ######################################################################

    _configButtonActions: dict[str, Callable[[], bool | None]] = {}
    _configUpdatedActions: dict[str, Callable[[Any], None]] = {}
    _config_help_actions: dict[str, Callable[[], str]] = {}

    def addonConfigDefaults(self, module: str) -> dict[str, Any] | None:
        path = os.path.join(self.addonsFolder(module), "config.json")
        try:
            with open(path, encoding="utf8") as f:
                return json.load(f)
        except Exception:
            return None

    def set_config_help_action(self, module: str, action: Callable[[], str]) -> None:
        "Set a callback used to produce config help."
        addon = self.addonFromModule(module)
        self._config_help_actions[addon] = action

    def addonConfigHelp(self, module: str) -> str:
        if action := self._config_help_actions.get(module, None):
            contents = action()
        else:
            path = os.path.join(self.addonsFolder(module), "config.md")
            if os.path.exists(path):
                with open(path, encoding="utf-8") as f:
                    contents = f.read()
            else:
                return ""

        return markdown.markdown(contents, extensions=[md_in_html.makeExtension()])

    def addonFromModule(self, module: str) -> str:  # softly deprecated
        return module.split(".")[0]

    @staticmethod
    def addon_from_module(module: str) -> str:
        return module.split(".")[0]

    def configAction(self, module: str) -> Callable[[], bool | None]:
        return self._configButtonActions.get(module)

    def configUpdatedAction(self, module: str) -> Callable[[Any], None]:
        return self._configUpdatedActions.get(module)

    # Schema
    ######################################################################

    def _addon_schema_path(self, module: str) -> str:
        return os.path.join(self.addonsFolder(module), "config.schema.json")

    def _addon_schema(self, module: str) -> Any:
        path = self._addon_schema_path(module)
        try:
            if not os.path.exists(path):
                # True is a schema accepting everything
                return True
            with open(path, encoding="utf-8") as f:
                return json.load(f)
        except json.decoder.JSONDecodeError as e:
            print("The schema is not valid:")
            print(e)

    # Add-on Config API
    ######################################################################

    def getConfig(self, module: str) -> dict[str, Any] | None:
        addon = self.addonFromModule(module)
        # get default config
        config = self.addonConfigDefaults(addon)
        if config is None:
            return None
        # merge in user's keys
        meta = self.addonMeta(addon)
        userConf = meta.get("config", {})
        config.update(userConf)
        return config

    def setConfigAction(self, module: str, fn: Callable[[], bool | None]) -> None:
        addon = self.addonFromModule(module)
        self._configButtonActions[addon] = fn

    def setConfigUpdatedAction(self, module: str, fn: Callable[[Any], None]) -> None:
        addon = self.addonFromModule(module)
        self._configUpdatedActions[addon] = fn

    def writeConfig(self, module: str, conf: dict) -> None:
        addon = self.addonFromModule(module)
        meta = self.addonMeta(addon)
        meta["config"] = conf
        self.writeAddonMeta(addon, meta)

    # user_files
    ######################################################################

    def _userFilesPath(self, sid: str) -> str:
        return os.path.join(self.addonsFolder(sid), "user_files")

    def _userFilesBackupPath(self) -> str:
        return os.path.join(self.addonsFolder(), "files_backup")

    def backupUserFiles(self, module: str) -> None:
        p = self._userFilesPath(module)

        if os.path.exists(p):
            os.rename(p, self._userFilesBackupPath())

    def restoreUserFiles(self, sid: str) -> None:
        p = self._userFilesPath(sid)
        bp = self._userFilesBackupPath()
        # did we back up userFiles?
        if not os.path.exists(bp):
            return
        os.rename(bp, p)

    # Web Exports
    ######################################################################

    _webExports: dict[str, str] = {}

    def setWebExports(self, module: str, pattern: str) -> None:
        addon = self.addonFromModule(module)
        self._webExports[addon] = pattern

    def getWebExports(self, module: str) -> str:
        return self._webExports.get(module)

    # Logging
    ######################################################################

    @classmethod
    def get_logger(cls, module: str) -> logging.Logger:
        """Return a logger for the given add-on module.

        NOTE: This method is static to allow it to be called outside of a
        running Anki instance, e.g. in add-on unit tests.
        """
        return logging.getLogger(
            f"{ADDON_LOGGER_PREFIX}{cls.addon_from_module(module)}"
        )

    def has_logger(self, module: str) -> bool:
        return find_addon_logger(self.addon_from_module(module)) is not None

    def is_debug_logging_enabled(self, module: str) -> bool:
        if not (logger := find_addon_logger(self.addon_from_module(module))):
            return False
        return logger.isEnabledFor(logging.DEBUG)

    def toggle_debug_logging(self, module: str, enable: bool) -> None:
        if not (logger := find_addon_logger(self.addon_from_module(module))):
            return
        logger.setLevel(logging.DEBUG if enable else logging.INFO)

    def logs_folder(self, module: str) -> Path:
        return get_addon_logs_folder(
            self.mw.pm.addon_logs(), self.addon_from_module(module)
        )


# Add-ons Dialog
######################################################################


class AddonsDialog(QDialog):
    def __init__(self, addonsManager: AddonManager) -> None:
        self.mgr = addonsManager
        self.mw = addonsManager.mw
        self._require_restart = False

        super().__init__(self.mw)

        f = self.form = aqt.forms.addons.Ui_Dialog()
        f.setupUi(self)
        qconnect(f.getAddons.clicked, self.onGetAddons)
        qconnect(f.installFromFile.clicked, self.onInstallFiles)
        qconnect(f.checkForUpdates.clicked, self.check_for_updates)
        qconnect(f.toggleEnabled.clicked, self.onToggleEnabled)
        qconnect(f.viewPage.clicked, self.onViewPage)
        qconnect(f.viewFiles.clicked, self.onViewFiles)
        qconnect(f.delete_2.clicked, self.onDelete)
        qconnect(f.config.clicked, self.onConfig)
        qconnect(self.form.addonList.itemDoubleClicked, self.onConfig)
        qconnect(self.form.addonList.currentRowChanged, self._onAddonItemSelected)
        qconnect(
            self.form.addonList.itemSelectionChanged, self._onAddonSelectionChanged
        )
        self.setWindowTitle(tr.addons_window_title())
        disable_help_button(self)
        self.setAcceptDrops(True)
        self.redrawAddons()
        restoreGeom(self, "addons")
        add_close_shortcut(self)
        gui_hooks.addons_dialog_will_show(self)
        self._onAddonSelectionChanged()
        self.show()

    def dragEnterEvent(self, event: QDragEnterEvent) -> None:
        mime = event.mimeData()
        if not mime.hasUrls():
            return None
        urls = mime.urls()
        exts = self.mgr.exts
        if all(any(url.toLocalFile().endswith(ext) for ext in exts) for url in urls):
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent) -> None:
        mime = event.mimeData()
        paths = []
        for url in mime.urls():
            path = url.toLocalFile()
            if os.path.exists(path):
                paths.append(path)
        self.onInstallFiles(paths)

    def reject(self) -> None:
        if self._require_restart:
            tooltip(tr.addons_changes_will_take_effect_when_anki(), parent=self.mw)
        saveGeom(self, "addons")
        aqt.dialogs.markClosed("AddonsDialog")

        return QDialog.reject(self)

    silentlyClose = True

    def name_for_addon_list(self, addon: AddonMeta) -> str:
        name = addon.human_name()

        if not addon.enabled:
            return f"{name} {tr.addons_disabled2()}"
        elif not addon.compatible():
            return f"{name} {tr.addons_requires(val=self.compatible_string(addon))}"

        return name

    def compatible_string(self, addon: AddonMeta) -> str:
        min = addon.min_version
        if min is not None and min > _current_version:
            ver = int_version_to_str(min)
            return f"Anki >= {ver}"
        else:
            max = abs(addon.max_version)
            ver = int_version_to_str(max)
            return f"Anki <= {ver}"

    def should_grey(self, addon: AddonMeta) -> bool:
        return not addon.enabled or not addon.compatible()

    def redrawAddons(
        self,
    ) -> None:
        addonList = self.form.addonList
        mgr = self.mgr

        self.addons = list(mgr.all_addon_meta())
        self.addons.sort(key=lambda a: a.human_name().lower())
        self.addons.sort(key=self.should_grey)

        selected = set(self.selectedAddons())
        addonList.clear()
        for addon in self.addons:
            name = self.name_for_addon_list(addon)
            item = QListWidgetItem(name, addonList)
            if self.should_grey(addon):
                item.setForeground(Qt.GlobalColor.gray)
            if addon.dir_name in selected:
                item.setSelected(True)

        addonList.reset()

    def _onAddonSelectionChanged(self) -> None:
        self.form.viewFiles.setEnabled(False)
        self.form.viewPage.setEnabled(False)
        self.form.config.setEnabled(False)

        selected_count = len(self.selectedAddons())
        if selected_count == 0:
            # View Files button shows top-level add-ons directory when nothing is selected
            self.form.viewFiles.setEnabled(True)
        elif selected_count == 1:
            addon = self.addons[self.form.addonList.currentRow()]

            self.form.viewFiles.setEnabled(True)
            self.form.viewPage.setEnabled(addon.page() is not None)
            self.form.config.setEnabled(
                bool(
                    self.mgr.getConfig(addon.dir_name)
                    or self.mgr.configAction(addon.dir_name)
                )
            )

    def _onAddonItemSelected(self, row_int: int) -> None:
        try:
            addon = self.addons[row_int]
        except IndexError:
            return
        gui_hooks.addons_dialog_did_change_selected_addon(self, addon)
        return

    def selectedAddons(self) -> list[str]:
        idxs = [x.row() for x in self.form.addonList.selectedIndexes()]
        return [self.addons[idx].dir_name for idx in idxs]

    def onlyOneSelected(self) -> str | None:
        dirs = self.selectedAddons()
        if len(dirs) != 1:
            show_info(tr.addons_please_select_a_single_addon_first())
            return None
        return dirs[0]

    def selected_addon_meta(self) -> AddonMeta | None:
        idxs = [x.row() for x in self.form.addonList.selectedIndexes()]
        if len(idxs) != 1:
            show_info(tr.addons_please_select_a_single_addon_first())
            return None
        return self.addons[idxs[0]]

    def onToggleEnabled(self) -> None:
        for module in self.selectedAddons():
            self.mgr.toggleEnabled(module)
        self._require_restart = True
        self.redrawAddons()

    def onViewPage(self) -> None:
        addon = self.selected_addon_meta()
        if not addon:
            return
        if page := addon.page():
            openLink(page)

    def onViewFiles(self) -> None:
        # if nothing selected, open top-level folder
        selected = self.selectedAddons()
        if not selected:
            openFolder(self.mgr.addonsFolder())
            return

        # otherwise require a single selection
        addon = self.onlyOneSelected()
        if not addon:
            return
        path = self.mgr.addonsFolder(addon)
        openFolder(path)

    def onDelete(self) -> None:
        selected = self.selectedAddons()
        if not selected:
            return
        if not askUser(tr.addons_delete_the_numd_selected_addon(count=len(selected))):
            return
        gui_hooks.addons_dialog_will_delete_addons(self, selected)
        try:
            for module in selected:
                # doing this before deleting, as `enabled` is always True afterwards
                if self.mgr.addon_meta(module).enabled:
                    self._require_restart = True
                self.mgr.deleteAddon(module)
        except OSError as e:
            showWarning(
                tr.addons_unable_to_update_or_delete_addon(val=str(e)),
                textFormat="plain",
            )
        self.form.addonList.clearSelection()
        self.redrawAddons()

    def onGetAddons(self) -> None:
        obj = GetAddons(self)
        if obj.ids:
            download_addons(
                self, self.mgr, obj.ids, self.after_downloading, force_enable=True
            )

    def after_downloading(self, log: list[DownloadLogEntry]) -> None:
        self.redrawAddons()
        if log:
            show_log_to_user(self, log)
        else:
            tooltip(tr.addons_no_updates_available())

    def onInstallFiles(self, paths: list[str] | None = None) -> bool | None:
        if not paths:
            filter = f"{tr.addons_packaged_anki_addon()} " + "({})".format(
                " ".join(f"*{ext}" for ext in self.mgr.exts)
            )
            paths_ = getFile(
                self, tr.addons_install_addons(), None, filter, key="addons", multi=True
            )
            paths = paths_  # type: ignore
            if not paths:
                return False

        installAddonPackages(self.mgr, paths, parent=self, force_enable=True)

        self.redrawAddons()
        return None

    def check_for_updates(self) -> None:
        tooltip(tr.addons_checking())
        check_and_prompt_for_updates(self, self.mgr, self.after_downloading)

    def onConfig(self) -> None:
        addon = self.onlyOneSelected()
        if not addon:
            return

        # does add-on manage its own config?
        act = self.mgr.configAction(addon)
        if act:
            ret = act()
            if ret is not False:
                return

        conf = self.mgr.getConfig(addon)
        if conf is None:
            showInfo(tr.addons_addon_has_no_configuration())
            return

        ConfigEditor(self, addon, conf)


# Fetching Add-ons
######################################################################


class GetAddons(QDialog):
    def __init__(self, dlg: AddonsDialog) -> None:
        QDialog.__init__(self, dlg)
        self.addonsDlg = dlg
        self.mgr = dlg.mgr
        self.mw = self.mgr.mw
        self.ids: list[int] = []
        self.form = aqt.forms.getaddons.Ui_Dialog()
        self.form.setupUi(self)
        b = self.form.buttonBox.addButton(
            tr.addons_browse_addons(), QDialogButtonBox.ButtonRole.ActionRole
        )
        qconnect(b.clicked, self.onBrowse)
        disable_help_button(self)
        restoreGeom(self, "getaddons", adjustSize=True)
        self.exec()
        saveGeom(self, "getaddons")

    def onBrowse(self) -> None:
        openLink(f"{aqt.appShared}addons/2.1")

    def accept(self) -> None:
        # get codes
        try:
            sids = self.form.code.text().split()
            sids = [
                re.sub(r"^https://ankiweb.net/shared/info/(\d+)$", r"\1", id_)
                for id_ in sids
            ]
            ids = [int(id_) for id_ in sids]
        except ValueError:
            showWarning(tr.addons_invalid_code())
            return

        self.ids = ids
        QDialog.accept(self)


# Downloading
######################################################################


def download_addon(client: HttpClient, id: int) -> DownloadOk | DownloadError:
    "Fetch a single add-on from AnkiWeb."
    try:
        resp = client.get(f"{aqt.appShared}download/{id}?v=2.1&p={_current_version}")
        if resp.status_code != 200:
            return DownloadError(status_code=resp.status_code)

        data = client.stream_content(resp)

        match = re.match(
            "attachment; filename=(.+)", resp.headers["content-disposition"]
        )
        assert match is not None
        fname = match.group(1)

        meta = extract_meta_from_download_url(resp.url)

        return DownloadOk(
            data=data,
            filename=fname,
            mod_time=meta.mod_time,
            min_point_version=meta.min_point_version,
            max_point_version=meta.max_point_version,
            branch_index=meta.branch_index,
        )
    except Exception as e:
        return DownloadError(exception=e)


@dataclass
class ExtractedDownloadMeta:
    mod_time: int
    min_point_version: int
    max_point_version: int
    branch_index: int


def extract_meta_from_download_url(url: str) -> ExtractedDownloadMeta:
    urlobj = urlparse(url)
    query = parse_qs(urlobj.query)

    def get_first_element(elements: list[str]) -> int:
        return int(elements[0])

    meta = ExtractedDownloadMeta(
        mod_time=get_first_element(query["t"]),
        min_point_version=get_first_element(query["minpt"]),
        max_point_version=get_first_element(query["maxpt"]),
        branch_index=get_first_element(query["bidx"]),
    )

    return meta


def download_log_to_html(log: list[DownloadLogEntry]) -> str:
    return "<br>".join(map(describe_log_entry, log))


def describe_log_entry(id_and_entry: DownloadLogEntry) -> str:
    (id, entry) = id_and_entry
    buf = f"{id}: "

    if isinstance(entry, DownloadError):
        if entry.status_code is not None:
            if entry.status_code in (403, 404):
                buf += tr.addons_invalid_code_or_addon_not_available()
            else:
                buf += tr.qt_misc_unexpected_response_code(val=entry.status_code)
        else:
            buf += (
                tr.addons_please_check_your_internet_connection()
                + "\n\n"
                + str(entry.exception)
            )
    elif isinstance(entry, InstallError):
        buf += entry.errmsg
    else:
        buf += tr.addons_installed_successfully()

    return buf


def download_encountered_problem(log: list[DownloadLogEntry]) -> bool:
    return any(not isinstance(e[1], InstallOk) for e in log)


def download_and_install_addon(
    mgr: AddonManager, client: HttpClient, id: int, force_enable: bool = False
) -> DownloadLogEntry:
    "Download and install a single add-on."
    result = download_addon(client, id)
    if isinstance(result, DownloadError):
        return (id, result)

    fname = result.filename.replace("_", " ")
    name = os.path.splitext(fname)[0].strip()
    if not name:
        name = str(id)

    manifest = dict(
        package=str(id),
        name=name,
        mod=result.mod_time,
        min_point_version=result.min_point_version,
        max_point_version=result.max_point_version,
        branch_index=result.branch_index,
    )

    result2 = mgr.install(
        io.BytesIO(result.data), manifest=manifest, force_enable=force_enable
    )

    return (id, result2)


class DownloaderInstaller(QObject):
    progressSignal = pyqtSignal(int, int)

    def __init__(self, parent: QWidget, mgr: AddonManager, client: HttpClient) -> None:
        QObject.__init__(self, parent)
        self.mgr = mgr
        self.client = client
        qconnect(self.progressSignal, self._progress_callback)

        def bg_thread_progress(up: int, down: int) -> None:
            self.progressSignal.emit(up, down)  # type: ignore

        self.client.progress_hook = bg_thread_progress

    def download(
        self,
        ids: list[int],
        on_done: Callable[[list[DownloadLogEntry]], None],
        force_enable: bool = False,
    ) -> None:
        self.ids = ids
        self.log: list[DownloadLogEntry] = []

        self.dl_bytes = 0
        self.last_tooltip = 0

        self.on_done = on_done

        parent = self.parent()
        assert isinstance(parent, QWidget)
        self.mgr.mw.progress.start(immediate=True, parent=parent)
        self.mgr.mw.taskman.run_in_background(
            lambda: self._download_all(force_enable), self._download_done
        )

    def _progress_callback(self, up: int, down: int) -> None:
        self.dl_bytes += down
        self.mgr.mw.progress.update(
            label=tr.addons_downloading_adbd_kb02fkb(
                part=len(self.log) + 1,
                total=len(self.ids),
                kilobytes=self.dl_bytes // 1024,
            )
        )

    def _download_all(self, force_enable: bool = False) -> None:
        for id in self.ids:
            self.log.append(
                download_and_install_addon(
                    self.mgr, self.client, id, force_enable=force_enable
                )
            )

    def _download_done(self, future: Future) -> None:
        self.mgr.mw.progress.finish()
        future.result()
        # qt gets confused if on_done() opens new windows while the progress
        # modal is still cleaning up
        self.mgr.mw.progress.single_shot(50, lambda: self.on_done(self.log))


def show_log_to_user(
    parent: QWidget, log: list[DownloadLogEntry], title: str = "Anki"
) -> None:
    have_problem = download_encountered_problem(log)

    if have_problem:
        text = tr.addons_one_or_more_errors_occurred()
    else:
        text = tr.addons_download_complete_please_restart_anki_to()
    text += f"<br><br>{download_log_to_html(log)}"

    if have_problem:
        showWarning(text, textFormat="rich", parent=parent, title=title)
    else:
        showInfo(text, parent=parent, title=title)


def download_addons(
    parent: QWidget,
    mgr: AddonManager,
    ids: list[int],
    on_done: Callable[[list[DownloadLogEntry]], None],
    client: HttpClient | None = None,
    force_enable: bool = False,
) -> None:
    if client is None:
        client = HttpClient()
    downloader = DownloaderInstaller(parent, mgr, client)
    downloader.download(ids, on_done=on_done, force_enable=force_enable)


# Update checking
######################################################################


class ChooseAddonsToUpdateList(QListWidget):
    ADDON_ID_ROLE = 101

    def __init__(
        self,
        parent: QWidget,
        mgr: AddonManager,
        updated_addons: list[AddonInfo],
    ) -> None:
        QListWidget.__init__(self, parent)
        self.mgr = mgr
        self.updated_addons = sorted(updated_addons, key=lambda addon: addon.modified)
        self.ignore_check_evt = False
        self.setup()
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        qconnect(self.itemClicked, self.on_click)
        qconnect(self.itemChanged, self.on_check)
        qconnect(self.itemDoubleClicked, self.on_double_click)
        qconnect(self.customContextMenuRequested, self.on_context_menu)

    def setup(self) -> None:
        header_item = QListWidgetItem(tr.addons_choose_update_update_all(), self)
        header_item.setFlags(
            Qt.ItemFlag(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
        )
        self.header_item = header_item
        for update_info in self.updated_addons:
            addon_id = update_info.id
            addon_meta = self.mgr.addon_meta(str(addon_id))
            update_enabled = addon_meta.update_enabled
            addon_name = addon_meta.human_name()
            update_timestamp = update_info.modified
            update_time = datetime.fromtimestamp(update_timestamp)

            addon_label = f"{update_time:%Y-%m-%d}   {addon_name}"
            item = QListWidgetItem(addon_label, self)
            # Not user checkable because it overlaps with itemClicked signal
            item.setFlags(Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled))
            if update_enabled:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            item.setData(self.ADDON_ID_ROLE, addon_id)
        self.refresh_header_check_state()

    def bool_to_check(self, check_bool: bool) -> Qt.CheckState:
        if check_bool:
            return Qt.CheckState.Checked
        else:
            return Qt.CheckState.Unchecked

    def checked(self, item: QListWidgetItem) -> bool:
        return item.checkState() == Qt.CheckState.Checked

    def on_click(self, item: QListWidgetItem) -> None:
        if item == self.header_item:
            return
        checked = self.checked(item)
        self.check_item(item, self.bool_to_check(not checked))
        self.refresh_header_check_state()

    def on_check(self, item: QListWidgetItem) -> None:
        if self.ignore_check_evt:
            return
        if item == self.header_item:
            self.header_checked(item.checkState())

    def on_double_click(self, item: QListWidgetItem) -> None:
        if item == self.header_item:
            checked = self.checked(item)
            self.check_item(self.header_item, self.bool_to_check(not checked))
            self.header_checked(self.bool_to_check(not checked))

    def on_context_menu(self, point: QPoint) -> None:
        if not (item := self.itemAt(point)):
            return
        addon_id = item.data(self.ADDON_ID_ROLE)
        m = QMenu()
        a = m.addAction(tr.addons_view_addon_page())
        qconnect(a.triggered, lambda _: openLink(f"{aqt.appShared}info/{addon_id}"))
        m.exec(QCursor.pos())

    def check_item(self, item: QListWidgetItem, check: Qt.CheckState) -> None:
        "call item.setCheckState without triggering on_check"
        self.ignore_check_evt = True
        item.setCheckState(check)
        self.ignore_check_evt = False

    def header_checked(self, check: Qt.CheckState) -> None:
        for i in range(1, self.count()):
            self.check_item(self.item(i), check)

    def refresh_header_check_state(self) -> None:
        for i in range(1, self.count()):
            item = self.item(i)
            if not self.checked(item):
                self.check_item(self.header_item, Qt.CheckState.Unchecked)
                return
        self.check_item(self.header_item, Qt.CheckState.Checked)

    def get_selected_addon_ids(self) -> list[int]:
        addon_ids = []
        for i in range(1, self.count()):
            item = self.item(i)
            if self.checked(item):
                addon_id = item.data(self.ADDON_ID_ROLE)
                addon_ids.append(addon_id)
        return addon_ids

    def save_check_state(self) -> None:
        for i in range(1, self.count()):
            item = self.item(i)
            addon_id = item.data(self.ADDON_ID_ROLE)
            addon_meta = self.mgr.addon_meta(str(addon_id))
            addon_meta.update_enabled = self.checked(item)
            self.mgr.write_addon_meta(addon_meta)


class ChooseAddonsToUpdateDialog(QDialog):
    _on_done: Callable[[list[int]], None]

    def __init__(
        self, parent: QWidget, mgr: AddonManager, updated_addons: list[AddonInfo]
    ) -> None:
        QDialog.__init__(self, parent)
        self.setWindowTitle(tr.addons_choose_update_window_title())
        self.setWindowModality(Qt.WindowModality.NonModal)
        self.mgr = mgr
        self.updated_addons = updated_addons
        self.setup()
        restoreGeom(self, "addonsChooseUpdate")

    def setup(self) -> None:
        layout = QVBoxLayout()
        label = QLabel(tr.addons_the_following_addons_have_updates_available())
        layout.addWidget(label)
        addons_list_widget = ChooseAddonsToUpdateList(
            self, self.mgr, self.updated_addons
        )
        layout.addWidget(addons_list_widget)
        self.addons_list_widget = addons_list_widget

        button_box = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )  # type: ignore
        qconnect(
            button_box.button(QDialogButtonBox.StandardButton.Ok).clicked, self.accept
        )
        qconnect(
            button_box.button(QDialogButtonBox.StandardButton.Cancel).clicked,
            self.reject,
        )
        layout.addWidget(button_box)
        self.setLayout(layout)

    def ask(self, on_done: Callable[[list[int]], None]) -> None:
        self._on_done = on_done
        self.show()

    def accept(self) -> None:
        saveGeom(self, "addonsChooseUpdate")
        self.addons_list_widget.save_check_state()
        self._on_done(self.addons_list_widget.get_selected_addon_ids())
        QDialog.accept(self)


def fetch_update_info(ids: list[int]) -> list[AddonInfo]:
    """Fetch update info from AnkiWeb in one or more batches."""
    all_info: list[AddonInfo] = []

    while ids:
        # get another chunk
        chunk = ids[:25]
        del ids[:25]

        batch_results = _fetch_update_info_batch(chunk)
        all_info.extend(batch_results)

    return all_info


def _fetch_update_info_batch(chunk: Iterable[int]) -> Sequence[AddonInfo]:
    return aqt.mw.backend.get_addon_info(
        client_version=_current_version, addon_ids=chunk
    )


def check_and_prompt_for_updates(
    parent: QWidget,
    mgr: AddonManager,
    on_done: Callable[[list[DownloadLogEntry]], None],
    requested_by_user: bool = True,
) -> None:
    def on_updates_received(items: list[AddonInfo]) -> None:
        handle_update_info(parent, mgr, items, on_done, requested_by_user)

    check_for_updates(mgr, on_updates_received)


def check_for_updates(
    mgr: AddonManager, on_done: Callable[[list[AddonInfo]], None]
) -> None:
    def check() -> list[AddonInfo]:
        return fetch_update_info(mgr.ankiweb_addons())

    def update_info_received(future: Future) -> None:
        # if syncing/in profile screen, defer message delivery
        if not mgr.mw.col:
            mgr.mw.progress.single_shot(
                1000,
                lambda: update_info_received(future),
                False,
            )
            return

        if future.exception():
            # swallow network errors
            print(str(future.exception()))
            result = []
        else:
            result = future.result()

        on_done(result)

    mgr.mw.taskman.run_in_background(check, update_info_received)


def handle_update_info(
    parent: QWidget,
    mgr: AddonManager,
    items: list[AddonInfo],
    on_done: Callable[[list[DownloadLogEntry]], None],
    requested_by_user: bool = True,
) -> None:
    mgr.update_supported_versions(items)
    updated_addons = mgr.get_updated_addons(items)

    if not updated_addons:
        on_done([])
        return

    prompt_to_update(parent, mgr, updated_addons, on_done, requested_by_user)


def prompt_to_update(
    parent: QWidget,
    mgr: AddonManager,
    updated_addons: list[AddonInfo],
    on_done: Callable[[list[DownloadLogEntry]], None],
    requested_by_user: bool = True,
) -> None:
    client = HttpClient()
    if not requested_by_user:
        prompt_update = False
        for addon in updated_addons:
            if mgr.addon_meta(str(addon.id)).update_enabled:
                prompt_update = True
        if not prompt_update:
            return

    def after_choosing(ids: list[int]) -> None:
        if ids:
            download_addons(parent, mgr, ids, on_done, client)

    ChooseAddonsToUpdateDialog(parent, mgr, updated_addons).ask(after_choosing)


def install_or_update_addon(
    parent: QWidget,
    mgr: AddonManager,
    addon_id: int,
    on_done: Callable[[list[DownloadLogEntry]], None],
) -> None:
    def check() -> list[AddonInfo]:
        return fetch_update_info([addon_id])

    def update_info_received(future: Future) -> None:
        try:
            items = future.result()
            updated_addons = mgr.get_updated_addons(items)
            if not updated_addons:
                on_done([])
                return
            client = HttpClient()
            download_addons(
                parent, mgr, [addon.id for addon in updated_addons], on_done, client
            )
        except Exception as exc:
            on_done([(addon_id, DownloadError(exception=exc))])

    mgr.mw.taskman.run_in_background(check, update_info_received)


# Editing config
######################################################################


class ConfigEditor(QDialog):
    def __init__(self, dlg: AddonsDialog, addon: str, conf: dict) -> None:
        super().__init__(dlg)
        self.addon = addon
        self.conf = conf
        self.mgr = dlg.mgr
        self.form = aqt.forms.addonconf.Ui_Dialog()
        self.form.setupUi(self)
        restore = self.form.buttonBox.button(
            QDialogButtonBox.StandardButton.RestoreDefaults
        )
        qconnect(restore.clicked, self.onRestoreDefaults)
        ok = self.form.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        ok.setShortcut(QKeySequence("Ctrl+Return"))
        self.setupFonts()
        self.updateHelp()
        self.updateText(self.conf)
        restoreGeom(self, "addonconf")
        self.form.splitter.setSizes([2 * self.width() // 3, self.width() // 3])
        restoreSplitter(self.form.splitter, "addonconf")
        self.setWindowTitle(
            without_unicode_isolation(
                tr.addons_config_window_title(
                    name=self.mgr.addon_meta(addon).human_name(),
                )
            )
        )
        disable_help_button(self)
        self.show()

    def onRestoreDefaults(self) -> None:
        default_conf = self.mgr.addonConfigDefaults(self.addon)
        self.updateText(default_conf)
        tooltip(tr.addons_restored_defaults(), parent=self)

    def setupFonts(self) -> None:
        font_mono = QFont("Consolas")
        if not font_mono.exactMatch():
            font_mono = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        font_mono.setPointSize(font_mono.pointSize())
        self.form.editor.setFont(font_mono)

    def updateHelp(self) -> None:
        txt = self.mgr.addonConfigHelp(self.addon)
        if txt:
            self.form.help.stdHtml(txt, js=[], css=["css/addonconf.css"], context=self)
        else:
            self.form.help.setVisible(False)

    def updateText(self, conf: dict[str, Any]) -> None:
        text = json.dumps(
            conf,
            ensure_ascii=False,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
        )
        text = gui_hooks.addon_config_editor_will_display_json(text)
        self.form.editor.setPlainText(text)
        if is_mac:
            self.form.editor.repaint()

    def onClose(self) -> None:
        saveGeom(self, "addonconf")
        saveSplitter(self.form.splitter, "addonconf")

    def reject(self) -> None:
        self.onClose()
        super().reject()

    def accept(self) -> None:
        txt = self.form.editor.toPlainText()
        txt = gui_hooks.addon_config_editor_will_update_json(txt, self.addon)
        try:
            new_conf = json.loads(txt)
            jsonschema.validate(new_conf, self.mgr._addon_schema(self.addon))
        except ValidationError as e:
            # The user did edit the configuration and entered a value
            # which can not be interpreted.
            schema = e.schema
            erroneous_conf = new_conf
            for link in e.path:
                erroneous_conf = erroneous_conf[link]
            path = "/".join(str(path) for path in e.path)
            if "error_msg" in schema:
                msg = schema["error_msg"].format(
                    problem=e.message,
                    path=path,
                    schema=str(schema),
                    erroneous_conf=erroneous_conf,
                )
            else:
                msg = tr.addons_config_validation_error(
                    problem=e.message,
                    path=path,
                    schema=str(schema),
                )
            showInfo(msg)
            return
        except Exception as e:
            showInfo(f"{tr.addons_invalid_configuration()} {repr(e)}")
            return

        if not isinstance(new_conf, dict):
            showInfo(tr.addons_invalid_configuration_top_level_object_must())
            return

        if new_conf != self.conf:
            self.mgr.writeConfig(self.addon, new_conf)
            # does the add-on define an action to be fired?
            act = self.mgr.configUpdatedAction(self.addon)
            if act:
                act(new_conf)

        self.onClose()
        super().accept()


# .ankiaddon installation wizard
######################################################################


def installAddonPackages(
    addonsManager: AddonManager,
    paths: list[str],
    parent: QWidget | None = None,
    warn: bool = False,
    strictly_modal: bool = False,
    advise_restart: bool = False,
    force_enable: bool = False,
) -> bool:
    if warn:
        names = ",<br>".join(f"<b>{os.path.basename(p)}</b>" for p in paths)
        q = tr.addons_important_as_addons_are_programs_downloaded() % dict(names=names)
        if (
            not showInfo(
                q,
                parent=parent,
                title=tr.addons_install_anki_addon(),
                type="warning",
                customBtns=[
                    QMessageBox.StandardButton.No,
                    QMessageBox.StandardButton.Yes,
                ],
            )
            == QMessageBox.StandardButton.Yes
        ):
            return False

    log, errs = addonsManager.processPackages(
        paths, parent=parent, force_enable=force_enable
    )

    if log:
        log_html = "<br>".join(log)
        if advise_restart:
            log_html += f"<br><br>{tr.addons_please_restart_anki_to_complete_the()}"
        if len(log) == 1 and not strictly_modal:
            tooltip(log_html, parent=parent)
        else:
            showInfo(
                log_html,
                parent=parent,
                textFormat="rich",
                title=tr.addons_installation_complete(),
            )
    if errs:
        msg = tr.addons_please_report_this_to_the_respective()
        showWarning(
            "<br><br>".join(errs + [msg]),
            parent=parent,
            textFormat="rich",
            title=tr.addons_addon_installation_error(),
        )

    return not errs



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

# ruff: noqa: F401
import sys

import aqt

from .browser import Browser, PreviewDialog

# aliases for legacy pathnames
from .sidebar import (
    SidebarItem,
    SidebarItemType,
    SidebarModel,
    SidebarSearchBar,
    SidebarStage,
    SidebarTool,
    SidebarToolbar,
    SidebarTreeView,
)
from .table import (
    CardState,
    Cell,
    CellRow,
    Column,
    Columns,
    DataModel,
    ItemId,
    ItemList,
    ItemState,
    NoteState,
    SearchContext,
    StatusDelegate,
    Table,
)

sys.modules["aqt.sidebar"] = sys.modules["aqt.browser.sidebar"]
aqt.sidebar = sys.modules["aqt.browser.sidebar"]  # type: ignore
sys.modules["aqt.previewer"] = sys.modules["aqt.browser.previewer"]
aqt.previewer = sys.modules["aqt.browser.previewer"]  # type: ignore
